import contextlib
import enum
import json
import os
import subprocess
import tempfile
import time
from pathlib import Path
from typing import Any, Optional, Set, Union

from osbuild.util import jsoncomm
from osbuild.util.fscache import FsCache, FsCacheInfo
from osbuild.util.mnt import mount, umount
from osbuild.util.path import clamp_mtime
from osbuild.util.types import PathLike

from . import api

__all__ = [
    "ObjectStore",
]


class PathAdapter:
    """Expose an object attribute as `os.PathLike`"""

    def __init__(self, obj: Any, attr: str) -> None:
        self.obj = obj
        self.attr = attr

    def __fspath__(self):
        return getattr(self.obj, self.attr)


class Object:
    class Mode(enum.IntEnum):
        READ = 0
        WRITE = 1

    class Metadata:
        """store and retrieve metadata for an object"""

        def __init__(self, base, folder: Optional[str] = None) -> None:
            self.base = base
            self.folder = folder
            os.makedirs(self.path, exist_ok=True)

        def _path_for_key(self, key) -> str:
            assert key
            name = f"{key}.json"
            return os.path.join(self.path, name)

        @property
        def path(self):
            if not self.folder:
                return self.base
            return os.path.join(self.base, self.folder)

        @contextlib.contextmanager
        def write(self, key):

            tmp = tempfile.NamedTemporaryFile(
                mode="w",
                encoding="utf8",
                dir=self.path,
                prefix=".",
                suffix=".tmp.json",
                delete=True,
            )

            with tmp as f:
                yield f

                f.flush()

                # if nothing was written to the file
                si = os.stat(tmp.name)
                if si.st_size == 0:
                    return

                dest = self._path_for_key(key)
                # ensure it is proper json?
                os.link(tmp.name, dest)

        @contextlib.contextmanager
        def read(self, key):
            dest = self._path_for_key(key)
            try:
                with open(dest, "r", encoding="utf8") as f:
                    yield f
            except FileNotFoundError:
                raise KeyError(f"No metadata for '{key}'") from None

        def set(self, key: str, data):

            if data is None:
                return

            with self.write(key) as f:
                json.dump(data, f, indent=2)

        def get(self, key: str):
            with contextlib.suppress(KeyError):
                with self.read(key) as f:
                    return json.load(f)
            return None

        def __fspath__(self):
            return self.path

    def __init__(self, cache: FsCache, uid: str, mode: Mode):
        self._cache = cache
        self._mode = mode
        self._id = uid
        self._path = None
        self._meta: Optional[Object.Metadata] = None
        self._stack: Optional[contextlib.ExitStack] = None
        self.source_epoch = None  # see finalize()

    def _open_for_reading(self):
        name = self._stack.enter_context(
            self._cache.load(self.id)
        )
        self._path = os.path.join(self._cache, name)

    def _open_for_writing(self):
        name = self._stack.enter_context(
            self._cache.stage()
        )
        self._path = os.path.join(self._cache, name)
        os.makedirs(os.path.join(self._path, "tree"))

    def to_dict(self):
        return {
            "mode": int(self._mode),
            "id": self._id,
            "path": str(Path(self._path).relative_to(self._cache)),
        }

    @classmethod
    def from_dict(cls, data, store):
        o = Object(store.cache, data["id"], Object.Mode(data["mode"]))
        o._path = Path(store.cache) / data["path"]
        o._stack = contextlib.ExitStack()
        return o

    def __enter__(self):
        assert not self.active
        self._stack = contextlib.ExitStack()
        if self.mode == Object.Mode.READ:
            self._open_for_reading()
        else:
            self._open_for_writing()

        # Expose our base path as `os.PathLike` via `PathAdater`
        # so any changes to it, e.g. via `store_tree`, will be
        # automatically picked up by `Metadata`.
        wrapped = PathAdapter(self, "_path")
        self._meta = self.Metadata(wrapped, folder="meta")

        if self.mode == Object.Mode.WRITE:
            self.meta.set("info", {
                "created": int(time.time()),
            })

        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        assert self.active
        self.cleanup()

    @property
    def active(self) -> bool:
        return self._stack is not None

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def mode(self) -> Mode:
        return self._mode

    def init(self, base: "Object"):
        """Initialize the object with the base object"""
        self._check_mode(Object.Mode.WRITE)
        assert self.active
        assert self._path

        subprocess.run(
            [
                "cp",
                "--reflink=auto",
                "-a",
                os.fspath(base.path) + "/.",
                os.fspath(self.path),
            ],
            check=True,
        )

    @property
    def path(self) -> str:
        assert self.active
        assert self._path
        return self._path

    @property
    def tree(self) -> str:
        return os.path.join(self.path, "tree")

    @property
    def meta(self) -> Metadata:
        assert self.active
        assert self._meta
        return self._meta

    @property
    def created(self) -> int:
        """When was the object created

        It is stored as `created` in the `info` metadata entry,
        and thus will also get overwritten if the metadata gets
        overwritten via `init()`.
        NB: only valid to access when the object is active.
        """
        info = self.meta.get("info")
        assert info, "info metadata missing"
        return info["created"]

    def clamp_mtime(self):
        """Clamp mtime of files and dirs to source_epoch

        If a source epoch is specified we clamp all files that
        are newer then our own creation timestap to the given
        source epoch. As a result all files created during the
        build should receive the source epoch modification time
        """
        if self.source_epoch is None:
            return

        clamp_mtime(self.tree, self.created, self.source_epoch)

    def finalize(self):
        if self.mode != Object.Mode.WRITE:
            return

        self.clamp_mtime()

        # put the object into the READER state
        self._mode = Object.Mode.READ

    def cleanup(self):
        if self._stack:
            self._stack.close()
            self._stack = None

    def _check_mode(self, want: Mode):
        """Internal: Raise a ValueError if we are not in the desired mode"""
        if self.mode != want:
            raise ValueError(f"Wrong object mode: {self.mode}, want {want}")

    def export(self, to_directory: PathLike, skip_preserve_owner=False):
        """Copy object into an external directory"""
        cp_cmd = [
            "cp",
            "--reflink=auto",
            "-a",
        ]
        if skip_preserve_owner:
            cp_cmd += ["--no-preserve=ownership"]
        cp_cmd += [
            os.fspath(self.tree) + "/.",
            os.fspath(to_directory),
        ]
        subprocess.run(cp_cmd, check=True)

    def __fspath__(self):
        return self.tree


class HostTree:
    """Read-only access to the host file system

    An object that provides the same interface as
    `objectstore.Object` that can be used to read
    the host file-system.
    """

    _root: Optional[tempfile.TemporaryDirectory]

    def __init__(self, store):
        self.store = store
        self._root = None
        self.init()

    def init(self):
        if self._root:
            return

        self._root = self.store.tempdir(prefix="host")

        root = self._root.name
        # Create a bare bones root file system. Starting with just
        # /usr mounted from the host.
        usr = os.path.join(root, "usr")
        os.makedirs(usr)
        # Also add in /etc/containers, which will allow us to access
        # /etc/containers/policy.json and enable moving containers
        # (skopeo): https://github.com/osbuild/osbuild/pull/1410
        # If https://github.com/containers/image/issues/2157 ever gets
        # fixed we can probably remove this bind mount.
        etc_containers = os.path.join(root, "etc", "containers")
        os.makedirs(etc_containers)

        # ensure / is read-only
        mount(root, root)
        mount("/usr", usr)
        mount("/etc/containers", etc_containers)

    @property
    def tree(self) -> os.PathLike:
        if not self._root:
            raise AssertionError("HostTree not initialized")
        return self._root.name

    def cleanup(self):
        if self._root:
            umount(self._root.name)
            self._root.cleanup()
            self._root = None

    def __fspath__(self) -> os.PathLike:
        return self.tree


class ObjectStore(contextlib.AbstractContextManager):
    def __init__(self, store: PathLike, read_only: bool = False):
        self.cache = FsCache("osbuild", store)
        self.tmp = os.path.join(store, "tmp")
        os.makedirs(self.store, exist_ok=True)
        os.makedirs(self.objects, exist_ok=True)
        os.makedirs(self.tmp, exist_ok=True)
        self._objs: Set[Object] = set()
        self._host_tree: Optional[HostTree] = None
        self._stack = contextlib.ExitStack()
        self._read_only = read_only

    def _get_floating(self, object_id: str) -> Optional[Object]:
        """Internal: get a non-committed object"""
        for obj in self._objs:
            if obj.mode == Object.Mode.READ and obj.id == object_id:
                return obj
        return None

    @property
    def maximum_size(self) -> Optional[Union[int, str]]:
        info = self.cache.info
        return info.maximum_size

    @maximum_size.setter
    def maximum_size(self, size: Union[int, str]):
        info = FsCacheInfo(maximum_size=size)
        self.cache.info = info

    @property
    def active(self) -> bool:
        # pylint: disable=protected-access
        return self.cache._is_active()

    @property
    def store(self):
        return os.fspath(self.cache)

    @property
    def objects(self):
        return os.path.join(self.cache, "objects")

    @property
    def host_tree(self) -> HostTree:
        assert self.active

        if not self._host_tree:
            self._host_tree = HostTree(self)
        return self._host_tree

    def contains(self, object_id, *, only_cached=False):
        if not object_id:
            return False

        if not only_cached and self._get_floating(object_id):
            return True

        try:
            with self.cache.load(object_id):
                return True
        except FsCache.MissError:
            return False

    def tempdir(self, prefix=None, suffix=None):
        """Return a tempfile.TemporaryDirectory within the store"""
        return tempfile.TemporaryDirectory(dir=self.tmp,
                                           prefix=prefix,
                                           suffix=suffix)

    def get(self, object_id):
        assert self.active

        obj = self._get_floating(object_id)
        if obj:
            return obj

        try:
            obj = Object(self.cache, object_id, Object.Mode.READ)
            self._stack.enter_context(obj)
            return obj
        except FsCache.MissError:
            return None

    def new(self, object_id: str):
        """Creates a new `Object` and open it for writing.

        It returns a instance of `Object` that can be used to
        write tree and metadata. Use `commit` to attempt to
        store the object in the cache.
        """
        assert self.active
        self._ensure_writable()

        obj = Object(self.cache, object_id, Object.Mode.WRITE)
        self._stack.enter_context(obj)

        self._objs.add(obj)

        return obj

    def commit(self, obj: Object, object_id: str):
        """Commits the Object to the object cache as `object_id`.

        Attempts to store the contents of `obj` and its metadata
        in the object cache. Whether anything is actually stored
        depends on the configuration of the cache, i.e. its size
        and how much free space is left or can be made available.
        Therefore the caller should not assume that the stored
        object can be retrived at all.
        """

        assert self.active
        self._ensure_writable()

        # we clamp the mtime of `obj` itself so that it
        # resuming a snapshop and building with a snapshot
        # goes through the same code path
        obj.clamp_mtime()

        self.cache.store_tree(object_id, obj.path + "/.")

    def cleanup(self):
        """Cleanup all created Objects that are still alive"""
        if self._host_tree:
            self._host_tree.cleanup()
            self._host_tree = None

        self._stack.close()
        self._objs = set()

    # export active floating objects so another store can load them,
    # typically in a VM
    def export_floating(self):
        floating = []
        for o in self._objs:
            if o.mode == Object.Mode.READ:
                floating.append(o.to_dict())
        return floating

    # Inject floating references from another store, used
    # to set up the ObjectStore in the VM corresponging
    # to the host one. Only valid for read-only stores.
    def load_floating(self, objs):
        if not self._read_only:
            raise RuntimeError("ObjectStore must be read-only to use load_floating()")
        self._objs = set()
        for o in objs:
            self._objs.add(Object.from_dict(o, self))

    def _ensure_writable(self):
        if self._read_only:
            raise RuntimeError("ObjectStore is not writable")

    def __fspath__(self):
        return os.fspath(self.store)

    def __enter__(self):
        assert not self.active
        self._stack.enter_context(self.cache)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert self.active
        self.cleanup()


class StoreServer(api.BaseAPI):

    endpoint = "store"

    def __init__(self, store: ObjectStore, *, socket_address=None):
        super().__init__(socket_address)
        self.store = store
        self.tmproot = store.tempdir(prefix="store-server-")
        self._stack = contextlib.ExitStack()

    def _cleanup(self):
        self.tmproot.cleanup()
        self.tmproot = None
        self._stack.close()
        self._stack = None

    def _read_tree(self, msg, sock):
        object_id = msg["object-id"]
        obj = self.store.get(object_id)
        if not obj:
            sock.send({"path": None})
            return

        sock.send({"path": obj.tree})

    def _read_tree_at(self, msg, sock):
        object_id = msg["object-id"]
        target = msg["target"]
        subtree = msg["subtree"]

        obj = self.store.get(object_id)
        if not obj:
            sock.send({"path": None})
            return

        try:
            source = os.path.join(obj, subtree.lstrip("/"))
            mount(source, target)
            self._stack.callback(umount, target)

        # pylint: disable=broad-except
        except Exception as e:
            sock.send({"error": str(e)})
            return

        sock.send({"path": target})

    def _mkdtemp(self, msg, sock):
        args = {
            "suffix": msg.get("suffix"),
            "prefix": msg.get("prefix"),
            "dir": self.tmproot.name
        }

        path = tempfile.mkdtemp(**args)
        sock.send({"path": path})

    def _source(self, msg, sock):
        name = msg["name"]
        base = self.store.store
        path = os.path.join(base, "sources", name)
        sock.send({"path": path})

    def _message(self, msg, _fds, sock):
        if msg["method"] == "read-tree":
            self._read_tree(msg, sock)
        elif msg["method"] == "read-tree-at":
            self._read_tree_at(msg, sock)
        elif msg["method"] == "mkdtemp":
            self._mkdtemp(msg, sock)
        elif msg["method"] == "source":
            self._source(msg, sock)
        else:
            raise ValueError("Invalid RPC call", msg)


class StoreClient:
    def __init__(self, connect_to="/run/osbuild/api/store"):
        self.client = jsoncomm.Socket.new_client(connect_to)

    def __del__(self):
        if self.client is not None:
            self.client.close()

    def mkdtemp(self, suffix=None, prefix=None):
        msg = {
            "method": "mkdtemp",
            "suffix": suffix,
            "prefix": prefix
        }

        self.client.send(msg)
        msg, _, _ = self.client.recv()

        return msg["path"]

    def read_tree(self, object_id: str):
        msg = {
            "method": "read-tree",
            "object-id": object_id
        }

        self.client.send(msg)
        msg, _, _ = self.client.recv()

        return msg["path"]

    def read_tree_at(self, object_id: str, target: str, path="/"):
        msg = {
            "method": "read-tree-at",
            "object-id": object_id,
            "target": os.fspath(target),
            "subtree": os.fspath(path)
        }

        self.client.send(msg)
        msg, _, _ = self.client.recv()

        err = msg.get("error")
        if err:
            raise RuntimeError(err)

        return msg["path"]

    def source(self, name: str) -> str:
        msg = {
            "method": "source",
            "name": name
        }

        self.client.send(msg)
        msg, _, _ = self.client.recv()

        return msg["path"]



"""Handling of experimental environment flags"""

import os
from typing import Any, Dict


def _experimental_env_map() -> Dict[str, Any]:
    env_map: Dict[str, Any] = {}
    for exp_opt in os.environ.get("OSBUILD_EXPERIMENTAL", "").split(","):
        l = exp_opt.split("=", maxsplit=1)
        if len(l) == 1:
            env_map[exp_opt] = "true"
        elif len(l) == 2:
            env_map[l[0]] = l[1]
    return env_map


def get_bool(option: str) -> bool:
    env_map = _experimental_env_map()
    opt = env_map.get(option, "")
    # sadly python as no strconv.ParseBool() like golang so we roll our own
    if opt.upper() in {"1", "T", "TRUE"}:
        return True
    if opt.upper() in {"", "0", "F", "FALSE"}:
        return False
    raise RuntimeError(f"unsupport bool val {opt}")


def get_string(option: str) -> str:
    env_map = _experimental_env_map()
    return str(env_map.get(option, ""))



# pylint: disable=fixme
# XXX: remove 'Provides: osbuild-dnf-json-api = 8' from the osbuild.spec file when this file is removed

import json
from typing import Any, Dict, TextIO

from osbuild.solver.exceptions import InvalidRequestError
from osbuild.solver.model import DepsolveResult, DumpResult, Package, Repository, SearchResult
from osbuild.solver.request import (
    DepsolveCmdArgs,
    DepsolveTransaction,
    SBOMRequest,
    SearchCmdArgs,
    SolverCommand,
    SolverConfig,
    SolverRequest,
)

from . import SolverAPIVersion


def _package_as_dict_dump_search(package: Package) -> dict:
    """
    Returns a dictionary representation of the RPM package for the DUMP and SEARCH commands.
    """
    return {
        "name": package.name,
        "summary": package.summary,
        "description": package.description,
        "url": package.url,
        "repo_id": package.repo_id,
        "epoch": package.epoch,
        "version": package.version,
        "release": package.release,
        "arch": package.arch,
        "buildtime": package.build_time_as_rfc3339(),
        "license": package.license,
    }


def _package_as_dict_depsolve(package: Package) -> dict:
    """
    Returns a dictionary representation of the RPM package for the DEPSOLVE command.
    """
    return {
        "name": package.name,
        "epoch": package.epoch,
        "version": package.version,
        "release": package.release,
        "arch": package.arch,
        "repo_id": package.repo_id,
        "path": package.location,
        "remote_location": package.remote_locations[0],
        "checksum": str(package.checksum),
    }


def _repository_as_dict(repository: Repository) -> dict:
    """
    Returns a dictionary representation of the repository. In v1, it is used only by the DEPSOLVE command.
    """
    return {
        "id": repository.repo_id,
        "name": repository.name,
        "baseurl": repository.baseurl,
        "metalink": repository.metalink,
        "mirrorlist": repository.mirrorlist,
        "gpgcheck": repository.gpgcheck,
        "repo_gpgcheck": repository.repo_gpgcheck,
        "gpgkeys": repository.gpgkey,
        "sslverify": repository.sslverify,
        "sslcacert": repository.sslcacert,
        "sslclientkey": repository.sslclientkey,
        "sslclientcert": repository.sslclientcert,
    }


# pylint: disable=unused-argument
def serialize_response_dump(solver: str, result: DumpResult, writer: TextIO) -> None:
    writer.write('[')
    first = True
    for package in result.packages:
        if not first:
            writer.write(', ')
        writer.write(json.dumps(_package_as_dict_dump_search(package)))
        first = False
    writer.write("]\n")


# pylint: disable=unused-argument
def serialize_response_search(solver: str, result: SearchResult, writer: TextIO) -> None:
    packages = [_package_as_dict_dump_search(package) for package in result.packages]
    writer.write(json.dumps(packages))
    writer.write("\n")


def serialize_response_depsolve(solver: str, result: DepsolveResult, writer: TextIO) -> None:
    last_transaction = result.transactions[-1] if result.transactions else []
    writer.write('{"solver": ')
    writer.write(json.dumps(solver))
    writer.write(', "packages": [')
    first = True
    for package in last_transaction:
        if not first:
            writer.write(', ')
        writer.write(json.dumps(_package_as_dict_depsolve(package)))
        first = False
    writer.write('], "repos": ')
    writer.write(json.dumps({repository.repo_id: _repository_as_dict(repository)
                 for repository in result.repositories}))
    writer.write(', "modules": ')
    writer.write(json.dumps(result.modules or {}))
    if result.sbom:
        writer.write(', "sbom": ')
        writer.write(json.dumps(result.sbom))
    writer.write('}\n')


# pylint: disable=too-many-branches
def _parse_repository(repo_dict: Dict[str, Any]) -> Repository:
    """
    Parse repository config from dict

    Note: Merges 'gpgkey' (singular) and 'gpgkeys' (plural) fields for backward compatibility.
    """

    if not isinstance(repo_dict, dict):
        raise TypeError("Repository config must be a dict")

    gpgkey = []
    # NB: 'gpgkey' is no longer used by osbuild/images implementation, remove it in the next API version.
    if "gpgkey" in repo_dict:
        gpgkey.append(repo_dict["gpgkey"])
    if "gpgkeys" in repo_dict:
        gpgkey.extend(repo_dict["gpgkeys"])

    kwargs = {
        "repo_id": repo_dict["id"],
    }
    if "name" in repo_dict:
        kwargs["name"] = repo_dict["name"]
    if "baseurl" in repo_dict:
        baseurl_value = repo_dict["baseurl"]
        if not isinstance(baseurl_value, list):
            raise TypeError(f"'baseurl' must be a list of URLs, got {type(baseurl_value).__name__}")
        kwargs["baseurl"] = baseurl_value
    if "metalink" in repo_dict:
        kwargs["metalink"] = repo_dict["metalink"]
    if "mirrorlist" in repo_dict:
        kwargs["mirrorlist"] = repo_dict["mirrorlist"]
    if "gpgcheck" in repo_dict:
        kwargs["gpgcheck"] = repo_dict["gpgcheck"]
    if "repo_gpgcheck" in repo_dict:
        kwargs["repo_gpgcheck"] = repo_dict["repo_gpgcheck"]
    if gpgkey:
        kwargs["gpgkey"] = gpgkey
    if "sslverify" in repo_dict:
        kwargs["sslverify"] = repo_dict["sslverify"]
    if "sslcacert" in repo_dict:
        kwargs["sslcacert"] = repo_dict["sslcacert"]
    if "sslclientkey" in repo_dict:
        kwargs["sslclientkey"] = repo_dict["sslclientkey"]
    if "sslclientcert" in repo_dict:
        kwargs["sslclientcert"] = repo_dict["sslclientcert"]
    if "metadata_expire" in repo_dict:
        kwargs["metadata_expire"] = repo_dict["metadata_expire"]
    if "module_hotfixes" in repo_dict:
        kwargs["module_hotfixes"] = repo_dict["module_hotfixes"]

    return Repository.from_request(**kwargs)


def _parse_depsolve_transaction(trans_dict: Dict[str, Any]) -> DepsolveTransaction:
    """Parse a transaction from a dict"""

    if not isinstance(trans_dict, dict):
        raise TypeError("Depsolve transaction must be a dict")

    kwargs = {
        "package_specs": trans_dict.get("package-specs", []),
    }
    if "exclude-specs" in trans_dict:
        kwargs["exclude_specs"] = trans_dict["exclude-specs"]
    if "repo-ids" in trans_dict:
        kwargs["repo_ids"] = trans_dict["repo-ids"]
    if "module-enable-specs" in trans_dict:
        kwargs["module_enable_specs"] = trans_dict["module-enable-specs"]
    if "install_weak_deps" in trans_dict:
        kwargs["install_weak_deps"] = trans_dict["install_weak_deps"]

    return DepsolveTransaction(**kwargs)


# pylint: disable=too-many-branches, too-many-statements
def parse_request(request_dict: Dict[str, Any]) -> SolverRequest:
    """Parse a V1 request dict into a SolverRequest object"""

    # Get required top-level fields
    try:
        command_str = request_dict["command"]
        arch = request_dict["arch"]
        releasever = request_dict["releasever"]
        cachedir = request_dict["cachedir"]
        arguments = request_dict["arguments"]
    except KeyError as e:
        raise InvalidRequestError(f"Missing required field {e}") from e

    # Validate and parse command
    try:
        command = SolverCommand(command_str)
    except ValueError as e:
        valid_cmds = ", ".join([c.value for c in SolverCommand])
        raise InvalidRequestError(
            f"Invalid command '{command_str}': must be one of {valid_cmds}"
        ) from e

    # Get optional top-level fields
    module_platform_id = request_dict.get("module_platform_id")
    proxy = request_dict.get("proxy")

    # Validate and parse command arguments
    if not isinstance(arguments, dict):
        raise InvalidRequestError("Field 'arguments' must be a dict")

    repos = None
    if "repos" in arguments:
        if not isinstance(arguments["repos"], list):
            raise InvalidRequestError("Field 'repos' must be a list")
        try:
            repos = [_parse_repository(r) for r in arguments["repos"]]
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid repository config: {e}") from e
        except KeyError as e:
            raise InvalidRequestError(f"Missing required field {e} in 'repos' item configuration") from e

    root_dir = arguments.get("root_dir")
    optional_metadata = arguments.get("optional-metadata")

    if optional_metadata and not isinstance(optional_metadata, list):
        raise InvalidRequestError("Field 'optional-metadata' must be a list")

    depsolve_transactions = []
    if command == SolverCommand.DEPSOLVE and "transactions" in arguments:
        if not isinstance(arguments["transactions"], list):
            raise InvalidRequestError("Field 'transactions' must be a list")
        try:
            depsolve_transactions = [_parse_depsolve_transaction(t) for t in arguments["transactions"]]
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid depsolve transaction: {e}") from e
        except KeyError as e:
            raise InvalidRequestError(f"Missing required field {e} in 'transactions' list") from e

    search_args = None
    if command == SolverCommand.SEARCH and "search" in arguments:
        if not isinstance(arguments["search"], dict):
            raise InvalidRequestError("Field 'search' must be a dict")
        try:
            packages = arguments["search"]["packages"]
            if not isinstance(packages, list):
                raise InvalidRequestError("Field 'packages' must be a list")
            latest = arguments["search"].get("latest", False)
            search_args = SearchCmdArgs(packages, latest)
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid search arguments: {e}") from e
        except KeyError as e:
            raise InvalidRequestError(f"Missing required field {e} in 'search' dict") from e

    sbom_request = None
    if "sbom" in arguments:
        if not isinstance(arguments["sbom"], dict):
            raise InvalidRequestError("Field 'sbom' must be a dict")
        try:
            sbom_request = SBOMRequest(arguments["sbom"]["type"])
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid value for 'type' in 'sbom': {e}") from e
        except KeyError as e:
            raise InvalidRequestError("Missing required field 'type' in 'sbom'") from e
        if command != SolverCommand.DEPSOLVE:
            raise InvalidRequestError("Field 'sbom' is only supported with 'depsolve' command")

    depsolve_args = None
    if depsolve_transactions:
        depsolve_args = DepsolveCmdArgs(
            transactions=depsolve_transactions,
            sbom_request=sbom_request,
        )

    return SolverRequest(
        api_version=SolverAPIVersion.V1,
        command=command,
        config=SolverConfig(
            arch=arch,
            releasever=releasever,
            cachedir=cachedir,
            module_platform_id=module_platform_id,
            proxy=proxy,
            repos=repos,
            root_dir=root_dir,
            optional_metadata=optional_metadata,
        ),
        depsolve_args=depsolve_args,
        search_args=search_args,
    )



#!/usr/bin/python3

from unittest.mock import call, patch

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.ostree.post-copy"


@patch("osbuild.util.ostree.cli")
def test_ostree_post_copy_smoke(mock_ostree_cli, stage_module):
    paths = {
        "mounts": "/run/osbuild/mounts",
    }
    options = {
        "sysroot": "/some/sysroot",
    }
    stage_module.main(paths, options)

    assert mock_ostree_cli.call_args_list == [
        call("admin", "post-copy", sysroot="/run/osbuild/mounts/some/sysroot")]


@pytest.mark.parametrize(
    "test_data,expected_err",
    [
        # devices is not used directly in the stage but it's required
        # because it's used as input for "mounts", see
        # https://github.com/osbuild/osbuild/pull/1343/files#r1402161208
        ({"devices": "must-be-object"}, " is not of type 'object'"),
        ({"mounts": "must-be-array"}, " is not of type 'array'"),
    ],
)
def test_schema_validation_ostree_post_copy(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
            "sysroot": "/some/sysroot",
        },
    }
    test_input.update(test_data)
    res = stage_schema.validate(test_input)

    assert res.valid is False
    testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)



#!/usr/bin/python3

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.dnf.module-config"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"conf": "must-be-object"}, "'must-be-object' is not of type 'object'"),
    # good
    ({
        "conf":
        {
            "name": "some-module",
            "stream": "some-stream",
            "state": "some-state",
            "profiles": ["some-profile"],
        },
    }, "")
])
def test_dnf_module_config_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


def test_dnf_module_config_writes_file(tmp_path, stage_module):
    treepath = tmp_path / "tree"
    confpath = "etc/dnf/modules.d/some-module.module"

    fullpath = treepath / confpath

    options = {
        "conf": {
            "name": "some-module",
            "stream": "some-stream",
            "state": "some-state",
            "profiles": ["some-profile", "other-profile"],
        }
    }

    stage_module.main(treepath, options)

    assert fullpath.exists()

    confdata = fullpath.read_text()

    assert confdata == """\
[some-module]
name = some-module
stream = some-stream
state = some-state
profiles = some-profile, other-profile

"""



"""
A base implementation of SPDX 2.3 model, as described on:
https://spdx.github.io/spdx-spec/v2.3/
"""

import hashlib
import re
from datetime import datetime, timezone
from enum import Enum, auto
from typing import Dict, List, Optional, Union


class CreatorType(Enum):
    """Enumeration of SPDX actor types."""

    PERSON = auto()
    ORGANIZATION = auto()
    TOOL = auto()

    def __str__(self) -> str:
        return self.name.capitalize()


class Creator():
    """Represents a Creator in SPDX."""

    def __init__(self, creator_type: CreatorType, name: str, email: Optional[str] = None) -> None:
        self.creator_type = creator_type
        self.name = name
        self.email = email

    def __str__(self):
        email_str = f" ({self.email})" if self.email else ""
        return f"{self.creator_type}: {self.name}{email_str}"


class EntityWithSpdxId():
    """
    Represents an SPDX entity with an SPDX ID.

    https://spdx.github.io/spdx-spec/v2.3/package-information/#72-package-spdx-identifier-field
    """

    def __init__(self, spdx_id: str) -> None:
        id_regex = re.compile(r"^SPDXRef-[a-zA-Z0-9\.\-]+$")
        if not id_regex.match(spdx_id):
            raise ValueError(f"Invalid SPDX ID '{spdx_id}'")
        self.spdx_id = spdx_id


def datetime_to_iso8601(dt: datetime) -> str:
    """
    Converts a datetime object to an SPDX-compliant ISO8601 string.

    This means that:
    - The timezone is UTC
    - The microsecond part is removed

    https://spdx.github.io/spdx-spec/v2.3/document-creation-information/#69-created-field
    """

    date = dt.astimezone(timezone.utc)
    date = date.replace(tzinfo=None)
    # Microseconds are not supported by SPDX
    date = date.replace(microsecond=0)
    return date.isoformat() + "Z"


class CreationInfo(EntityWithSpdxId):
    """
    Represents SPDX creation information.

    https://spdx.github.io/spdx-spec/v2.3/document-creation-information/
    """

    def __init__(
        self,
        spdx_version: str,
        spdx_id: str,
        name: str,
        document_namespace: str,
        creators: List[Creator],
        created: datetime,
        data_license: str = "CC0-1.0",
    ) -> None:
        super().__init__(spdx_id)

        if not spdx_version.startswith("SPDX-"):
            raise ValueError(f"Invalid SPDX version '{spdx_version}'")

        if spdx_id != "SPDXRef-DOCUMENT":
            raise ValueError(f"Invalid SPDX ID '{spdx_id}'")

        self.spdx_version = spdx_version
        self.name = name
        self.data_license = data_license
        self.document_namespace = document_namespace
        self.creators = creators
        self.created = created

    def to_dict(self):
        return {
            "SPDXID": self.spdx_id,
            "creationInfo": {
                "created": datetime_to_iso8601(self.created),
                "creators": [str(creator) for creator in self.creators],
            },
            "dataLicense": self.data_license,
            "name": self.name,
            "spdxVersion": self.spdx_version,
            "documentNamespace": self.document_namespace,
        }


class NoAssertionValue():
    """Represents the SPDX No Assertion value."""

    VALUE = "NOASSERTION"

    def __str__(self):
        return self.VALUE


class NoneValue():
    """Represents the SPDX None value."""

    VALUE = "NONE"

    def __str__(self):
        return self.VALUE


class ExternalPackageRefCategory(Enum):
    """Enumeration of external package reference categories."""

    SECURITY = auto()
    PACKAGE_MANAGER = auto()
    PERSISTENT_ID = auto()
    OTHER = auto()

    def __str__(self) -> str:
        return self.name.replace("_", "-")


CATEGORY_TO_REPOSITORY_TYPE: Dict[ExternalPackageRefCategory, List[str]] = {
    ExternalPackageRefCategory.SECURITY: ["cpe22Type", "cpe23Type", "advisory", "fix", "url", "swid"],
    ExternalPackageRefCategory.PACKAGE_MANAGER: ["maven-central", "nuget", "bower", "purl"],
    ExternalPackageRefCategory.PERSISTENT_ID: ["swh", "gitoid"],
    ExternalPackageRefCategory.OTHER: [],
}


class ExternalPackageRef():
    """
    Represents an external package reference.

    https://spdx.github.io/spdx-spec/v2.3/package-information/#721-external-reference-field
    """

    def __init__(self, category: ExternalPackageRefCategory, reference_type: str, locator: str) -> None:
        if len(CATEGORY_TO_REPOSITORY_TYPE[category]
               ) > 0 and reference_type not in CATEGORY_TO_REPOSITORY_TYPE[category]:
            raise ValueError(f"Invalid repository type '{reference_type}' for category '{category}'")

        self.category = category
        self.reference_type = reference_type
        self.locator = locator

    def to_dict(self):
        return {
            "referenceCategory": str(self.category),
            "referenceType": self.reference_type,
            "referenceLocator": self.locator,
        }


class ChecksumAlgorithm(Enum):
    """Enumeration of SPDX checksum algorithms."""

    SHA1 = auto()
    SHA224 = auto()
    SHA256 = auto()
    SHA384 = auto()
    SHA512 = auto()
    SHA3_256 = auto()
    SHA3_384 = auto()
    SHA3_512 = auto()
    BLAKE2b_256 = auto()
    BLAKE2b_384 = auto()
    BLAKE2b_512 = auto()
    BLAKE3 = auto()
    MD2 = auto()
    MD4 = auto()
    MD5 = auto()
    MD6 = auto()
    ADLER32 = auto()

    def __str__(self) -> str:
        return self.name.replace("_", "-")


class Checksum():
    """
    Represents a checksum.

    https://spdx.github.io/spdx-spec/v2.3/package-information/#72-checksum-fields
    """

    def __init__(self, algorithm: ChecksumAlgorithm, value: str) -> None:
        self.algorithm = algorithm
        self.value = value

    def to_dict(self):
        return {
            "algorithm": str(self.algorithm),
            "checksumValue": self.value,
        }


def normalize_name_for_license_id(name: str) -> str:
    """
    Normalize a license name to be used within an SPDX license ID.

    The function does the following things:
    - Ensures that the returned string contains only letters, numbers, "." and/or "-".
        All other characters are replaced with "-".
    - Deduplicates consecutive "." and "-" characters.

    See also:
    https://spdx.github.io/spdx-spec/v2.3/other-licensing-information-detected/#1011-description:
    """
    normalized_name = re.sub(r"[^a-zA-Z0-9.-]", "-", name)
    normalized_name = re.sub(r"([.-])\1+", r"\1", normalized_name)
    return normalized_name


def generate_license_id(extracted_text: str, name: Optional[str] = None) -> str:
    """
    Generate a unique SPDX license ID by hashing the extracted text using SHA-256.

    If a license name is provided, include it in the license ID.
    """
    extracted_text_hash = hashlib.sha256(extracted_text.encode()).hexdigest()
    if name is not None:
        return f"LicenseRef-{normalize_name_for_license_id(name)}-{extracted_text_hash}"
    return f"LicenseRef-{extracted_text_hash}"


class ExtractedLicensingInfo():
    """
    Represents extracted licensing information for a license not on the SPDX License List.

    https://spdx.github.io/spdx-spec/v2.3/other-licensing-information-detected/
    """

    def __init__(self, extracted_text: str, name: Optional[str] = None) -> None:
        self.extracted_text = extracted_text
        self.name = name
        self.license_ref_id = generate_license_id(self.extracted_text, self.name)

    def __str__(self):
        return self.license_ref_id

    def to_dict(self):
        d = {
            "licenseId": self.license_ref_id,
            "extractedText": self.extracted_text,
        }
        if self.name:
            d["name"] = self.name
        return d


# pylint: disable=too-many-instance-attributes
class Package(EntityWithSpdxId):
    """Represents an SPDX package."""

    def __init__(
        self,
        spdx_id: str,
        name: str,
        download_location: Union[str, NoAssertionValue, NoneValue],
        version: Optional[str] = None,
        files_analyzed: Optional[bool] = None,
        checksums: Optional[List[Checksum]] = None,
        homepage: Optional[Union[str, NoAssertionValue, NoneValue]] = None,
        source_info: Optional[str] = None,
        license_declared: Optional[Union[str, ExtractedLicensingInfo, NoAssertionValue, NoneValue]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        external_references: Optional[List[ExternalPackageRef]] = None,
        built_date: Optional[datetime] = None,
    ) -> None:
        super().__init__(spdx_id)
        self.name = name
        self.download_location = download_location
        self.version = version
        self.files_analyzed = files_analyzed
        self.checksums = checksums or []
        self.homepage = homepage
        self.source_info = source_info
        self.license_declared = license_declared
        self.summary = summary
        self.description = description
        self.external_references = external_references or []
        self.built_date = built_date

    def to_dict(self):
        d = {
            "SPDXID": self.spdx_id,
            "name": self.name,
            "downloadLocation": str(self.download_location)
        }
        if self.files_analyzed is not None:
            d["filesAnalyzed"] = self.files_analyzed
        if self.version:
            d["versionInfo"] = self.version
        if self.checksums:
            d["checksums"] = [checksum.to_dict() for checksum in self.checksums]
        if self.homepage:
            d["homepage"] = str(self.homepage)
        if self.source_info:
            d["sourceInfo"] = self.source_info
        if self.license_declared:
            d["licenseDeclared"] = str(self.license_declared)
        if self.summary:
            d["summary"] = self.summary
        if self.description:
            d["description"] = self.description
        if self.external_references:
            d["externalRefs"] = [ref.to_dict() for ref in self.external_references]
        if self.built_date:
            d["builtDate"] = datetime_to_iso8601(self.built_date)
        return d


class RelationshipType(Enum):
    """Enumeration of SPDX relationship types."""

    DESCRIBES = auto()
    DEPENDS_ON = auto()
    OPTIONAL_DEPENDENCY_OF = auto()

    def __str__(self) -> str:
        return self.name


class Relationship():
    """Represents a relationship between SPDX elements."""

    def __init__(
        self,
        spdx_element_id: str,
        relationship_type: RelationshipType,
        related_spdx_element_id: Union[str, NoneValue, NoAssertionValue],
        comment: Optional[str] = None,
    ) -> None:
        self.spdx_element_id = spdx_element_id
        self.relationship_type = relationship_type
        self.related_spdx_element_id = related_spdx_element_id
        self.comment = comment

    def to_dict(self):
        d = {
            "spdxElementId": self.spdx_element_id,
            "relationshipType": str(self.relationship_type),
            "relatedSpdxElement": str(self.related_spdx_element_id),
        }
        if self.comment:
            d["comment"] = self.comment
        return d


class Document():
    """Represents an SPDX document."""

    def __init__(
        self,
        creation_info: CreationInfo,
        packages: Optional[List[Package]] = None,
        relationships: Optional[List[Relationship]] = None,
        extracted_licensing_infos: Optional[List[ExtractedLicensingInfo]] = None,
    ) -> None:
        self.creation_info = creation_info
        self.packages = packages or []
        self.relationships = relationships or []
        self.extracted_licensing_infos = extracted_licensing_infos or []

    def to_dict(self):
        d = self.creation_info.to_dict()
        for package in self.packages:
            d.setdefault("packages", []).append(package.to_dict())
        for extracted_licensing_info in self.extracted_licensing_infos:
            d.setdefault("hasExtractedLicensingInfos", []).append(extracted_licensing_info.to_dict())
        for relationship in self.relationships:
            d.setdefault("relationships", []).append(relationship.to_dict())
        return d



#!/usr/bin/python3
import re

import pytest

from osbuild.testutil import (
    assert_jsonschema_error_contains,
)

STAGE_NAME = "org.osbuild.treeinfo"

# https://release-engineering.github.io/productmd/treeinfo-1.0.html#examples
treeinfo_1_0_example_output = """
[checksums]
images/boot.iso=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
images/efiboot.img=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
images/macboot.img=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
images/product.img=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
images/pxeboot/initrd.img=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
images/pxeboot/upgrade.img=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
images/pxeboot/vmlinuz=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
repodata/repomd.xml=sha256:e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855

[general]
arch=x86_64
family=Fedora
name=Fedora 21
packagedir=Packages
platforms=x86_64,xen
repository=.
timestamp=1417653911
variant=Server
version=21

[images-x86_64]
boot.iso=images/boot.iso
initrd=images/pxeboot/initrd.img
kernel=images/pxeboot/vmlinuz
upgrade=images/pxeboot/upgrade.img

[images-xen]
initrd=images/pxeboot/initrd.img
kernel=images/pxeboot/vmlinuz
upgrade=images/pxeboot/upgrade.img

[release]
name=Fedora
short=Fedora
version=21

[stage2]
mainimage=LiveOS/squashfs.img

[tree]
arch=x86_64
build_timestamp=1417653911
platforms=x86_64,xen
variants=Server

[variant-Server]
id=Server
name=Server
packages=Packages
repository=.
type=variant
uid=Server

[header]
version=1.0
"""

treeinfo_1_0_example_input = {
    "path": "/tmp/.treeinfo",
    "treeinfo": {
        "checksums": [
            "images/boot.iso",
            "images/efiboot.img",
            "images/macboot.img",
            "images/product.img",
            "images/pxeboot/initrd.img",
            "images/pxeboot/upgrade.img",
            "images/pxeboot/vmlinuz",
            "repodata/repomd.xml"
        ],
        "general": {
            "arch": "x86_64",
            "family": "Fedora",
            "name": "Fedora 21",
            "packagedir": "Packages",
            "platforms": [
                "x86_64",
                "xen"
            ],
            "repository": ".",
            "timestamp": 1417653911,
            "variant": "Server",
            "version": "21"
        },
        "images-x86_64": {
            "boot.iso": "images/boot.iso",
            "initrd": "images/pxeboot/initrd.img",
            "kernel": "images/pxeboot/vmlinuz",
            "upgrade": "images/pxeboot/upgrade.img"
        },
        "images-xen": {
            "initrd": "images/pxeboot/initrd.img",
            "kernel": "images/pxeboot/vmlinuz",
            "upgrade": "images/pxeboot/upgrade.img"
        },
        "release": {
            "name": "Fedora",
            "short": "Fedora",
            "version": "21"
        },
        "stage2": {
            "mainimage": "LiveOS/squashfs.img"
        },
        "tree": {
            "arch": "x86_64",
            "build_timestamp": 1417653911,
            "platforms": [
                "x86_64",
                "xen"
            ],
            "variants": [
                "Server"
            ]
        },
        "variant-Server": {
            "id": "Server",
            "name": "Server",
            "packages": "Packages",
            "repository": ".",
            "type": "variant",
            "uid": "Server"
        }
    }
}


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({"path": ""}, r"'treeinfo' is a required property"),
    ({"treeinfo": {}}, r"'path' is a required property"),
    # good
    (treeinfo_1_0_example_input, ""),
])
def test_treeinfo_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, re.compile(expected_err), expected_num_errs=1)


def test_treeinfo_writes_file(tmp_path, stage_module):
    treepath = tmp_path / "tree"
    confpath = "tmp/.treeinfo"
    fullpath = treepath / confpath

    # mock function make_sum to return known sha256 for empty files
    def mock_make_sum(_: str, file_path: str):
        return file_path, "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
    stage_module.make_sum = mock_make_sum

    stage_module.main(treepath, treeinfo_1_0_example_input)

    assert fullpath.exists()

    confdata = fullpath.read_text()

    assert "\n" + confdata == treeinfo_1_0_example_output + "\n"



import os
import pathlib
import subprocess
from types import ModuleType

import pytest

import osbuild.meta
from osbuild.testutil.imports import import_module_from_path


@pytest.fixture
def stage_module(request: pytest.FixtureRequest) -> ModuleType:
    """stage_module is a fixture that imports a stage module by its name
    defined in STAGE_NAME in the test module.
    """
    if not hasattr(request.module, "STAGE_NAME"):
        raise ValueError("stage_module fixture must be used in a test module that defines STAGE_NAME")

    stage_name = request.module.STAGE_NAME
    caller_dir = pathlib.Path(request.node.fspath).parent
    module_path = caller_dir.parent / stage_name
    return import_module_from_path("stage", os.fspath(module_path))


@pytest.fixture
def stage_schema(request: pytest.FixtureRequest) -> osbuild.meta.Schema:
    """stage_schema is a fixture returns the schema for a stage module.

    This fixture may be indirectly parametrized with the stage schema version.
    If the schema version is not specified, the version "2" is assumed.
    The stage name must be defined in STAGE_NAME in the test module.
    """
    if hasattr(request, "param") and not isinstance(request.param, str):
        raise ValueError(
            "stage_schema fixture may be indirectly parametrized only with the stage schema version string")

    if not hasattr(request.module, "STAGE_NAME"):
        raise ValueError("stage_schema fixture must be used in a test module that defines STAGE_NAME")

    stage_name = request.module.STAGE_NAME
    schema_version = request.param if hasattr(request, "param") else "2"
    caller_dir = pathlib.Path(request.node.fspath).parent
    root = caller_dir.parent.parent
    mod_info = osbuild.meta.ModuleInfo.load(root, "Stage", stage_name)
    return osbuild.meta.Schema(mod_info.get_schema(version=schema_version), stage_name)


@pytest.fixture
def tmp_path_disk_full(tmp_path, tmp_path_disk_full_size):
    small_dir = tmp_path / "small-dir"
    small_dir.mkdir()
    subprocess.run(["mount",
                    "-t", "tmpfs",
                    "-o", f"size={tmp_path_disk_full_size}",
                    "tmpfs",
                    small_dir
                    ], check=True)
    yield small_dir
    subprocess.run(["umount", small_dir], check=True)



"""
Device Handling for pipeline stages

Specific type of artifacts require device support, such as
loopback devices or device mapper. Since stages are always
run in a container and are isolated from the host, they do
not have direct access to devices and specifically can not
setup new ones.
Therefore device handling is done at the osbuild level with
the help of a device host services. Device specific modules
provide the actual functionality and thus the core device
support in osbuild itself is abstract.
"""

import abc
import errno
import hashlib
import json
import os
import stat
from pathlib import Path
from typing import Any, Dict, Optional

from osbuild import host
from osbuild.mixins import MixinImmutableID
from osbuild.util import ctx


class Device(MixinImmutableID):
    """
    A single device with its corresponding options
    """

    def __init__(self, name, info, parent, options: Dict):
        self.name = name
        self.info_name = info.name
        self.info_path = info.path
        self.parent = parent
        self.options = options or {}
        self.id = self.calc_id()

    def calc_id(self):
        # NB: Since the name of the device is arbitrary or prescribed
        # by the stage, it is not included in the id calculation.
        m = hashlib.sha256()

        m.update(json.dumps(self.info_name, sort_keys=True).encode())
        if self.parent:
            m.update(json.dumps(self.parent.id, sort_keys=True).encode())
        m.update(json.dumps(self.options, sort_keys=True).encode())
        return m.hexdigest()

    def to_dict(self, libdir) -> Dict:
        return {
            "name": self.name,
            "info_name": self.info_name,
            "info_path": str(Path(self.info_path).relative_to(libdir)),
            "parent_name": self.parent.name if self.parent else None,
            "options": self.options,
        }

    @classmethod
    def from_dict(cls, data: Dict, libdir, parent: Optional["Device"] = None) -> "Device":
        class InfoStub:
            def __init__(self, name: str, path: str):
                self.name = name
                self.path = path

        info = InfoStub(data["info_name"], Path(libdir) / data["info_path"])
        dev = cls(
            name=data["name"],
            info=info,
            parent=parent,
            options=data["options"]
        )
        return dev


class DeviceManager:
    """Manager for Devices

    Uses a `host.ServiceManager` to open `Device` instances.
    """

    def __init__(self, mgr: host.ServiceManager, devpath: str, tree_path: str) -> None:
        self.service_manager = mgr
        self.devpath = devpath
        self.tree = tree_path
        self.devices: Dict[str, Dict[str, Any]] = {}

    def device_relpath(self, dev: Optional[Device]) -> Optional[str]:
        if dev is None:
            return None
        return self.devices[dev.name]["path"]

    def device_abspath(self, dev: Optional[Device]) -> Optional[str]:
        relpath = self.device_relpath(dev)
        if relpath is None:
            return None
        return os.path.join(self.devpath, relpath)

    def open(self, dev: Device) -> Dict:

        parent = self.device_relpath(dev.parent)

        args = {
            # global options
            "dev": self.devpath,
            "tree": self.tree,

            "parent": parent,

            # per device options
            "options": dev.options,
        }

        mgr = self.service_manager

        client = mgr.start(f"device/{dev.name}", dev.info_path)
        res = client.call("open", args)

        self.devices[dev.name] = res
        return res


class DeviceService(host.Service):
    """Device host service"""

    @staticmethod
    def ensure_device_node(path, major: int, minor: int, dir_fd=None):
        """Ensure that the specified device node exists at the given path"""
        mode = 0o666 | stat.S_IFBLK
        with ctx.suppress_oserror(errno.EEXIST):
            os.mknod(path, mode, os.makedev(major, minor), dir_fd=dir_fd)

    @abc.abstractmethod
    def open(self, devpath: str, parent: str, tree: str, options: Dict):
        """Open a specific device

        This method must be implemented by the specific device service.
        It should open the device and create a device node in `devpath`.
        The return value must contain the relative path to the device
        node.
        """

    @abc.abstractmethod
    def close(self):
        """Close the device"""

    def stop(self):
        self.close()

    def dispatch(self, method: str, args, _fds):
        if method == "open":
            r = self.open(args["dev"],
                          args["parent"],
                          args["tree"],
                          args["options"])
            return r, None
        if method == "close":
            r = self.close()
            return r, None

        raise host.ProtocolError("Unknown method")



# V2 API implementation
#
# For schema consistency, all fields are always included, even if they are not
# set or set to empty values.
#
# The only exception is the 'sbom' field, which is only included if the 'sbom'
# was explicitly requested.

import json
from typing import Any, Dict, List, Set, TextIO

from osbuild.solver.exceptions import InvalidRequestError
from osbuild.solver.model import (
    Checksum,
    Dependency,
    DepsolveResult,
    DumpResult,
    Package,
    Repository,
    SearchResult,
)
from osbuild.solver.request import (
    DepsolveCmdArgs,
    DepsolveTransaction,
    SBOMRequest,
    SearchCmdArgs,
    SolverCommand,
    SolverConfig,
    SolverRequest,
)

from . import SolverAPIVersion


def _checksum_as_dict(checksum: Checksum) -> Dict[str, Any]:
    return {
        "algorithm": checksum.algorithm,
        "value": str(checksum.value),
    }


def _dependency_as_dict(dependency: Dependency) -> Dict[str, Any]:
    d = {"name": dependency.name}
    if dependency.relation:
        d["relation"] = dependency.relation
    if dependency.version:
        d["version"] = dependency.version
    return d


def _package_as_dict(package: Package) -> Dict[str, Any]:
    """
    Returns a dictionary representation of the RPM package.

    All fields are always included for schema consistency.
    Scalar fields may be None, list fields are empty arrays if not set.
    """
    return {
        # Core fields (always expected to have values)
        "name": package.name,
        "epoch": package.epoch,
        "version": package.version,
        "release": package.release,
        "arch": package.arch,
        "repo_id": package.repo_id,
        "location": package.location,
        "remote_locations": package.remote_locations,
        "checksum": _checksum_as_dict(package.checksum) if package.checksum else None,

        # Metadata fields (may be None)
        "header_checksum": _checksum_as_dict(package.header_checksum) if package.header_checksum else None,
        "license": package.license,
        "summary": package.summary,
        "description": package.description,
        "url": package.url,
        "vendor": package.vendor,
        "packager": package.packager,
        "build_time": package.build_time_as_rfc3339() if package.build_time else None,
        "download_size": package.download_size,
        "install_size": package.install_size,
        "group": package.group,
        "source_rpm": package.source_rpm,
        "reason": package.reason,

        # Dependency lists (always arrays, may be empty)
        "provides": [_dependency_as_dict(dep) for dep in package.provides],
        "requires": [_dependency_as_dict(dep) for dep in package.requires],
        "requires_pre": [_dependency_as_dict(dep) for dep in package.requires_pre],
        "conflicts": [_dependency_as_dict(dep) for dep in package.conflicts],
        "obsoletes": [_dependency_as_dict(dep) for dep in package.obsoletes],
        "regular_requires": [_dependency_as_dict(dep) for dep in package.regular_requires],
        "recommends": [_dependency_as_dict(dep) for dep in package.recommends],
        "suggests": [_dependency_as_dict(dep) for dep in package.suggests],
        "enhances": [_dependency_as_dict(dep) for dep in package.enhances],
        "supplements": [_dependency_as_dict(dep) for dep in package.supplements],

        # File list (always array, may be empty)
        "files": package.files,
    }


def _repository_as_dict(repository: Repository) -> Dict[str, Any]:
    """
    Returns a dictionary representation of the repository.

    All fields from the Repository model are included for schema consistency.

    When repository.rhsm=True, SSL secret fields (sslcacert, sslclientkey, sslclientcert)
    are set to None as they contain host-specific RHSM secrets that should not be exposed.
    """
    d = {
        "id": repository.repo_id,
        "name": repository.name,
        "baseurl": repository.baseurl,
        "metalink": repository.metalink,
        "mirrorlist": repository.mirrorlist,
        "gpgcheck": repository.gpgcheck,
        "repo_gpgcheck": repository.repo_gpgcheck,
        "gpgkey": repository.gpgkey,
        "sslverify": repository.sslverify,
        "metadata_expire": repository.metadata_expire,
        "module_hotfixes": repository.module_hotfixes,
        "rhsm": repository.rhsm,
        # SSL secrets are set to None when using RHSM (host-specific secrets not applicable),
        # otherwise return the actual values from DNF (which may be empty strings if not configured)
        "sslcacert": None if repository.rhsm else repository.sslcacert,
        "sslclientkey": None if repository.rhsm else repository.sslclientkey,
        "sslclientcert": None if repository.rhsm else repository.sslclientcert,
    }
    return d


def serialize_response_dump(solver: str, result: DumpResult, writer: TextIO) -> None:
    writer.write('{"solver": ')
    writer.write(json.dumps(solver))
    writer.write(', "packages": [')
    first = True
    for package in result.packages:
        if not first:
            writer.write(', ')
        writer.write(json.dumps(_package_as_dict(package)))
        first = False
    writer.write('], "repos": ')
    writer.write(json.dumps({r.repo_id: _repository_as_dict(r) for r in result.repositories}))
    writer.write('}\n')


def serialize_response_search(solver: str, result: SearchResult, writer: TextIO) -> None:
    d = {
        "solver": solver,
        "packages": [_package_as_dict(package) for package in result.packages],
        "repos": {repository.repo_id: _repository_as_dict(repository) for repository in result.repositories},
    }
    writer.write(json.dumps(d))
    writer.write("\n")


def _transactions_to_disjoint_sets(transactions: List[List[Package]]) -> List[List[Package]]:
    """
    Convert a list of transactions to a list of disjoint sets of packages.

    Solver implementations always return transactions as supersets of the previous transaction results.
    This function converts the transactions to a list of disjoint sets of packages, where each transaction result contains
    only the new "to be installed" packages.
    """
    disjoint_sets: List[List[Package]] = []
    seen_packages: Set[Package] = set()
    for transaction in transactions:
        current_set = set(transaction)
        disjoint_transaction = list(current_set - seen_packages)
        seen_packages.update(disjoint_transaction)
        # NOTE: we sort the transaction to ensure that the transaction results are
        # kept sorted in the same way as the original transactions.
        disjoint_transaction.sort()
        disjoint_sets.append(disjoint_transaction)
    return disjoint_sets


def serialize_response_depsolve(solver: str, result: DepsolveResult, writer: TextIO) -> None:
    """
    Serializes a Solver API response for the DEPSOLVE command.

    The response is designed to be convenient for the osbuild/images implementation.
    Specifically, the response will be used to generate multiple instances of the org.osbuild.rpm stage
    in a manifest to install the packages in groups in the same order as the transactions.
    For this reason, we do not return the transactions results as supersets of the previous transaction results,
    but rather post-process the data to generate transactions results, where each transaction result contains
    only the new "to be installed" packages. Effectively making the transactions results disjoint sets of packages.

    Example response:
    {
        "solver": "<solver name>",
        "transactions": [
            [
                {
                    "name": "bash",
                    "epoch": 0,
                    "version": "5.1.8",
                    "release": "9.el9",
                    "arch": "x86_64",
                    "repo_id": "fedora",
                    "location": "Packages/bash-5.1.8-9.el9.x86_64.rpm",
                    "remote_locations": ["https://example.com/fedora/Packages/bash-5.1.8-9.el9.x86_64.rpm"],
                    "checksum": "sha256:aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                    "license": "GPLv2+",
                    "summary": "The GNU Bourne Again shell",
                    "description": "The GNU Bourne Again shell",
                    ...
                },
                {
                    "name": "glibc",
                    "epoch": 0,
                    "version": "2.33",
                    "release": "1.fc33",
                    "arch": "x86_64",
                    "repo_id": "fedora",
                    "location": "Packages/glibc-2.33-1.fc33.x86_64.rpm",
                    "remote_locations": ["https://example.com/fedora/Packages/glibc-2.33-1.fc33.x86_64.rpm"],
                    "checksum": "sha256:bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
                    "license": "LGPLv2+",
                    "summary": "The GNU C Library",
                    "description": "The GNU C Library",
                    ...
                }
            ],
            [
                {
                    "name": "libstdc++",
                    "epoch": 0,
                    "version": "11.0.1",
                    "release": "1.fc33",
                    "arch": "x86_64",
                    "repo_id": "fedora",
                    "location": "Packages/libstdc++-11.0.1-1.fc33.x86_64.rpm",
                    "remote_locations": ["https://example.com/fedora/Packages/libstdc++-11.0.1-1.fc33.x86_64.rpm"],
                    "checksum": "sha256:cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc",
                    "license": "GPLv2+",
                    "summary": "The GNU Standard C++ Library",
                    "description": "The GNU Standard C++ Library",
                    ...
                }
            ],
            ...
        ],
        "repos": {
            "fedora": {
                "id": "fedora",
                "name": "Fedora",
                "baseurl": ["https://example.com/fedora"],
            }
        },
        "modules": {...},
        "sbom": {...},
    }
    """
    transactions = _transactions_to_disjoint_sets(result.transactions)
    writer.write('{"solver": ')
    writer.write(json.dumps(solver))
    writer.write(', "transactions": [')
    first = True
    for transaction in transactions:
        if not first:
            writer.write(', ')
        writer.write(json.dumps([_package_as_dict(package) for package in transaction]))
        first = False
    writer.write('], "repos": ')
    writer.write(json.dumps({repository.repo_id: _repository_as_dict(repository)
                 for repository in result.repositories}))
    writer.write(', "modules": ')
    writer.write(json.dumps(result.modules or {}))
    if result.sbom:
        writer.write(', "sbom": ')
        writer.write(json.dumps(result.sbom))
    writer.write('}\n')


# pylint: disable=too-many-branches
def _parse_repository(repo_dict: Dict[str, Any]) -> Repository:
    """
    Parse repository config from dict.

    Note: V2 API only supports 'gpgkey' as a list. Unlike V1, it does not support
    'gpgkey' as a string or merge 'gpgkey' and 'gpgkeys' fields.
    """

    if not isinstance(repo_dict, dict):
        raise TypeError("Repository config must be a dict")

    kwargs = {
        "repo_id": repo_dict["id"],
    }
    if "name" in repo_dict:
        kwargs["name"] = repo_dict["name"]
    if "baseurl" in repo_dict:
        baseurl_value = repo_dict["baseurl"]
        if not isinstance(baseurl_value, list):
            raise TypeError(f"'baseurl' must be a list of URLs, got {type(baseurl_value).__name__}")
        kwargs["baseurl"] = baseurl_value
    if "metalink" in repo_dict:
        kwargs["metalink"] = repo_dict["metalink"]
    if "mirrorlist" in repo_dict:
        kwargs["mirrorlist"] = repo_dict["mirrorlist"]
    if "gpgcheck" in repo_dict:
        kwargs["gpgcheck"] = repo_dict["gpgcheck"]
    if "repo_gpgcheck" in repo_dict:
        kwargs["repo_gpgcheck"] = repo_dict["repo_gpgcheck"]
    if "gpgkey" in repo_dict:
        gpgkey_value = repo_dict["gpgkey"]
        if not isinstance(gpgkey_value, list):
            raise TypeError(f"'gpgkey' must be a list, got {type(gpgkey_value).__name__}")
        kwargs["gpgkey"] = gpgkey_value
    if "sslverify" in repo_dict:
        kwargs["sslverify"] = repo_dict["sslverify"]
    if "sslcacert" in repo_dict:
        kwargs["sslcacert"] = repo_dict["sslcacert"]
    if "sslclientkey" in repo_dict:
        kwargs["sslclientkey"] = repo_dict["sslclientkey"]
    if "sslclientcert" in repo_dict:
        kwargs["sslclientcert"] = repo_dict["sslclientcert"]
    if "metadata_expire" in repo_dict:
        kwargs["metadata_expire"] = repo_dict["metadata_expire"]
    if "module_hotfixes" in repo_dict:
        kwargs["module_hotfixes"] = repo_dict["module_hotfixes"]
    if "rhsm" in repo_dict:
        kwargs["rhsm"] = repo_dict["rhsm"]
    if "enabled" in repo_dict:
        kwargs["enabled"] = repo_dict["enabled"]
    if "priority" in repo_dict:
        kwargs["priority"] = repo_dict["priority"]
    if "cost" in repo_dict:
        kwargs["cost"] = repo_dict["cost"]
    if "username" in repo_dict:
        kwargs["username"] = repo_dict["username"]
    if "password" in repo_dict:
        kwargs["password"] = repo_dict["password"]
    if "skip_if_unavailable" in repo_dict:
        kwargs["skip_if_unavailable"] = repo_dict["skip_if_unavailable"]

    return Repository.from_request(**kwargs)


def _parse_depsolve_transaction(trans_dict: Dict[str, Any]) -> DepsolveTransaction:
    """Parse a transaction from a dict"""

    if not isinstance(trans_dict, dict):
        raise TypeError("Depsolve transaction must be a dict")

    kwargs = {
        "package_specs": trans_dict.get("package-specs", []),
    }

    if not kwargs["package_specs"]:
        raise ValueError("Depsolve transaction must contain at least one package specification")

    if "exclude-specs" in trans_dict:
        kwargs["exclude_specs"] = trans_dict["exclude-specs"]
    if "repo-ids" in trans_dict:
        kwargs["repo_ids"] = trans_dict["repo-ids"]
    if "module-enable-specs" in trans_dict:
        kwargs["module_enable_specs"] = trans_dict["module-enable-specs"]
    if "install_weak_deps" in trans_dict:
        kwargs["install_weak_deps"] = trans_dict["install_weak_deps"]

    return DepsolveTransaction(**kwargs)


# pylint: disable=too-many-branches, too-many-statements
def parse_request(request_dict: Dict[str, Any]) -> SolverRequest:
    """Parse a V2 request dict into a SolverRequest object"""

    # Get required top-level fields
    try:
        command_str = request_dict["command"]
        arch = request_dict["arch"]
        releasever = request_dict["releasever"]
        cachedir = request_dict["cachedir"]
        arguments = request_dict["arguments"]
    except KeyError as e:
        raise InvalidRequestError(f"Missing required field {e}") from e

    # Validate and parse command
    try:
        command = SolverCommand(command_str)
    except ValueError as e:
        valid_cmds = ", ".join([c.value for c in SolverCommand])
        raise InvalidRequestError(
            f"Invalid command '{command_str}': must be one of {valid_cmds}"
        ) from e

    # Get optional top-level fields
    module_platform_id = request_dict.get("module_platform_id")
    proxy = request_dict.get("proxy")

    # Validate and parse command arguments
    if not isinstance(arguments, dict):
        raise InvalidRequestError("Field 'arguments' must be a dict")

    repos = None
    if "repos" in arguments:
        if not isinstance(arguments["repos"], list):
            raise InvalidRequestError("Field 'repos' must be a list")
        try:
            repos = [_parse_repository(r) for r in arguments["repos"]]
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid repository config: {e}") from e
        except KeyError as e:
            raise InvalidRequestError(f"Missing required field {e} in 'repos' item configuration") from e

    root_dir = arguments.get("root_dir")
    optional_metadata = arguments.get("optional-metadata")

    if optional_metadata and not isinstance(optional_metadata, list):
        raise InvalidRequestError("Field 'optional-metadata' must be a list")

    depsolve_transactions = []
    if "transactions" in arguments:
        if not isinstance(arguments["transactions"], list):
            raise InvalidRequestError("Field 'transactions' must be a list")
        try:
            depsolve_transactions = [_parse_depsolve_transaction(t) for t in arguments["transactions"]]
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid depsolve transaction: {e}") from e
        except KeyError as e:
            raise InvalidRequestError(f"Missing required field {e} in 'transactions' list") from e

    search_args = None
    if "search" in arguments:
        if not isinstance(arguments["search"], dict):
            raise InvalidRequestError("Field 'search' must be a dict")
        try:
            packages = arguments["search"]["packages"]
            if not isinstance(packages, list):
                raise InvalidRequestError("Field 'packages' must be a list")
            latest = arguments["search"].get("latest", False)
            search_args = SearchCmdArgs(packages, latest)
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid search arguments: {e}") from e
        except KeyError as e:
            raise InvalidRequestError(f"Missing required field {e} in 'search' dict") from e

    sbom_request = None
    if "sbom" in arguments:
        if not isinstance(arguments["sbom"], dict):
            raise InvalidRequestError("Field 'sbom' must be a dict")
        try:
            sbom_request = SBOMRequest(arguments["sbom"]["type"])
        except (ValueError, TypeError) as e:
            raise InvalidRequestError(f"Invalid value for 'type' in 'sbom': {e}") from e
        except KeyError as e:
            raise InvalidRequestError("Missing required field 'type' in 'sbom'") from e
        if command != SolverCommand.DEPSOLVE:
            raise InvalidRequestError("Field 'sbom' is only supported with 'depsolve' command")

    depsolve_args = None
    if depsolve_transactions:
        depsolve_args = DepsolveCmdArgs(
            transactions=depsolve_transactions,
            sbom_request=sbom_request,
        )

    return SolverRequest(
        api_version=SolverAPIVersion.V2,
        command=command,
        config=SolverConfig(
            arch=arch,
            releasever=releasever,
            cachedir=cachedir,
            module_platform_id=module_platform_id,
            proxy=proxy,
            repos=repos,
            root_dir=root_dir,
            optional_metadata=optional_metadata,
        ),
        depsolve_args=depsolve_args,
        search_args=search_args,
    )



from datetime import datetime
from typing import Dict, List

import dnf
import hawkey

import osbuild.util.sbom.model as sbom_model


def bom_chksum_algorithm_from_hawkey(chksum_type: int) -> sbom_model.ChecksumAlgorithm:
    """
    Convert a hawkey checksum type number to an SBOM checksum algorithm.
    """
    if chksum_type == hawkey.CHKSUM_MD5:
        return sbom_model.ChecksumAlgorithm.MD5
    if chksum_type == hawkey.CHKSUM_SHA1:
        return sbom_model.ChecksumAlgorithm.SHA1
    if chksum_type == hawkey.CHKSUM_SHA256:
        return sbom_model.ChecksumAlgorithm.SHA256
    if chksum_type == hawkey.CHKSUM_SHA384:
        return sbom_model.ChecksumAlgorithm.SHA384
    if chksum_type == hawkey.CHKSUM_SHA512:
        return sbom_model.ChecksumAlgorithm.SHA512
    raise ValueError(f"Unknown Hawkey checksum type: {chksum_type}")


def _hawkey_reldep_to_rpmdependency(reldep: hawkey.Reldep) -> sbom_model.RPMDependency:
    """
    Convert a hawkey.Reldep to an SBOM RPM dependency.
    """
    try:
        return sbom_model.RPMDependency(reldep.name, reldep.relation, reldep.version)
    except AttributeError:
        # '_hawkey.Reldep' object has no attribute 'name' in the version shipped on RHEL-8
        dep_parts = str(reldep).split()
        while len(dep_parts) < 3:
            dep_parts.append("")
        return sbom_model.RPMDependency(dep_parts[0], dep_parts[1], dep_parts[2])


# pylint: disable=too-many-branches
def dnf_pkgset_to_sbom_pkgset(dnf_pkgset: List[dnf.package.Package]) -> List[sbom_model.BasePackage]:
    """
    Convert a dnf package set to a SBOM package set.
    """
    pkgs_by_name = {}
    pkgs_by_provides: Dict[str, List[sbom_model.BasePackage]] = {}

    for dnf_pkg in dnf_pkgset:
        pkg = sbom_model.RPMPackage(
            name=dnf_pkg.name,
            version=dnf_pkg.version,
            release=dnf_pkg.release,
            architecture=dnf_pkg.arch,
            epoch=dnf_pkg.epoch,
            license_declared=dnf_pkg.license,
            vendor=dnf_pkg.vendor,
            build_date=datetime.fromtimestamp(dnf_pkg.buildtime),
            summary=dnf_pkg.summary,
            description=dnf_pkg.description,
            source_rpm=dnf_pkg.sourcerpm,
            homepage=dnf_pkg.url,
        )

        if dnf_pkg.chksum:
            pkg.checksums = {
                bom_chksum_algorithm_from_hawkey(dnf_pkg.chksum[0]): dnf_pkg.chksum[1].hex()
            }

        if dnf_pkg.remote_location():
            pkg.download_url = dnf_pkg.remote_location()

        # if dnf_pkg.from_repo is empty, the pkg is not installed. determine from remote_location
        # if dnf_pkg.from_repo is "@commanddline", the pkg was installed from the command line, there is no repo URL
        # if dnf_pkg.reponame is "@System", the package is installed and there is no repo URL
        # if dnf_pkg.from_repo is a string with repo ID, determine the repo URL from the repo configuration
        if not dnf_pkg.from_repo and dnf_pkg.remote_location():
            pkg.repository_url = dnf_pkg.remote_location()[:-len("/" + dnf_pkg.relativepath)]
        elif dnf_pkg.from_repo != "@commandline" and dnf_pkg.reponame != "@System":
            repo_url = ""
            if dnf_pkg.repo.baseurl:
                repo_url = dnf_pkg.repo.baseurl
            elif dnf_pkg.repo.metalink:
                repo_url = dnf_pkg.repo.metalink
            elif dnf_pkg.repo.mirrorlist:
                repo_url = dnf_pkg.repo.mirrorlist
            pkg.repository_url = repo_url

        pkg.rpm_provides = [_hawkey_reldep_to_rpmdependency(r) for r in dnf_pkg.provides]
        pkg.rpm_requires = [_hawkey_reldep_to_rpmdependency(r) for r in dnf_pkg.requires]
        pkg.rpm_recommends = [_hawkey_reldep_to_rpmdependency(r) for r in dnf_pkg.recommends]
        pkg.rpm_suggests = [_hawkey_reldep_to_rpmdependency(r) for r in dnf_pkg.suggests]

        # The dnf_pkgset is not sorted by package dependencies. We need to determine relationships in two steps:
        # 1. Collect all packages that provide a certain capability
        # 2. Resolve dependencies for each package using previously constructed list of capabilities by package.
        # Doing this in two steps ensures that all soft dependencies satisfied by a package from the same set are
        # resolved.
        for provide in pkg.rpm_provides:
            pkgs_by_provides.setdefault(provide.name, []).append(pkg)
        # Packages can also depend directly on files provided by other packages. Collect these as well.
        for provided_file in dnf_pkg.files:
            pkgs_by_provides.setdefault(provided_file, []).append(pkg)

        pkgs_by_name[pkg.name] = pkg

    for pkg in pkgs_by_name.values():
        for require in pkg.rpm_requires:
            # skip conditional dependencies if the required package is not in the set
            # "relation" contains whitespace on both sides
            if require.relation.strip() == "if" and pkgs_by_name.get(require.version) is None:
                continue
            for provider_pkg in pkgs_by_provides.get(require.name, []):
                pkg.depends_on.add(provider_pkg)

        for soft_dep in pkg.rpm_recommends + pkg.rpm_suggests:
            for provider_pkg in pkgs_by_provides.get(soft_dep.name, []):
                pkg.optional_depends_on.add(provider_pkg)

    return list(pkgs_by_name.values())



#!/usr/bin/python3

import os.path
import re
import subprocess

import pytest

from osbuild import testutil
from osbuild.testutil import has_executable

TEST_INPUT = [
    ({"lang": "en_US.UTF-8"}, "lang en_US.UTF-8"),
    ({"keyboard": "us"}, "keyboard us"),
    ({"timezone": "UTC"}, "timezone UTC"),
    (
        {
            "lang": "en_US.UTF-8",
            "keyboard": "us",
            "timezone": "UTC",
        },
        "lang en_US.UTF-8\nkeyboard us\ntimezone UTC",
    ),
    ({"rootpw": {"lock": True}}, "rootpw --lock"),
    ({"rootpw": {"plaintext": True, "password": "plaintext-password"}}, "rootpw --plaintext plaintext-password"),
    ({"rootpw": {"iscrypted": True, "password": "encrypted-password"}}, "rootpw --iscrypted encrypted-password"),
    (
        {"rootpw": {"iscrypted": True, "allow_ssh": True, "password": "encrypted-password"}},
        "rootpw --iscrypted --allow-ssh encrypted-password",
    ),
    (
        {"rootpw": {"plaintext": True, "allow_ssh": True, "password": "plaintext-password"}},
        "rootpw --plaintext --allow-ssh plaintext-password",
    ),
    (
        {"rootpw": {"plaintext": True, "lock": True, "password": "plaintext-password"}},
        "rootpw --lock --plaintext plaintext-password",
    ),
    (
        {
            "ostree": {
                "osname": "some-osname",
                "url": "http://some-ostree-url.com/foo",
                "ref": "some-ref",
                "remote": "some-remote",
                "gpg": True,
            },
            "liveimg": {
                "url": "some-liveimg-url",
            },
            "groups": {
                "somegrp": {
                    "gid": 2337,
                },
                "dot.group": {},
            },
            "users": {
                "someusr": {
                    "uid": 1337,
                    "gid": 1337,
                    "groups": [
                        "grp1",
                        "grp2",
                    ],
                    "home": "/other/home/someusr",
                    "shell": "/bin/ksh",
                    "password": "$1$notreally",
                    "key": "ssh-rsa not-really-a-real-key",
                },
                "dot.user": {
                    "key": "ssh-rsa not-really-a-real-key",
                },
            },
        },
        "ostreesetup --osname=some-osname --url=http://some-ostree-url.com/foo --ref=some-ref --remote=some-remote\n"
        + "liveimg --url some-liveimg-url\n"
        + "group --name somegrp --gid 2337\n"
        + "group --name dot.group\n"
        + "user --name someusr --password $1$notreally --iscrypted --shell /bin/ksh --uid 1337 --gid 1337 --groups grp1,grp2 --homedir /other/home/someusr\n"
        + 'sshkey --username someusr "ssh-rsa not-really-a-real-key"\n'
        + "user --name dot.user\n"
        + 'sshkey --username dot.user "ssh-rsa not-really-a-real-key"',
    ),
    ({"zerombr": True}, "zerombr"),
    ({"clearpart": {"all": True}}, "clearpart --all"),
    ({"clearpart": {"all": True, "initlabel": True}}, "clearpart --all --initlabel"),
    (
        {"clearpart": {"drives": ["sd*|hd*|vda", "/dev/vdc"]}},
        "clearpart --drives=sd*|hd*|vda,/dev/vdc",
    ),
    ({"clearpart": {"drives": ["hda"]}}, "clearpart --drives=hda"),
    (
        {"clearpart": {"drives": ["disk/by-id/scsi-58095BEC5510947BE8C0360F604351918"]}},
        "clearpart --drives=disk/by-id/scsi-58095BEC5510947BE8C0360F604351918"
    ),
    ({"clearpart": {"drives": ["hda"], "initlabel": True}}, "clearpart --drives=hda --initlabel"),
    ({"clearpart": {"list": ["sda2", "sda3"]}}, "clearpart --list=sda2,sda3"),
    ({"clearpart": {"list": ["sda2"]}}, "clearpart --list=sda2"),
    (
        {"clearpart": {"disklabel": "some-label"}},
        "clearpart --disklabel=some-label",
    ),
    ({"clearpart": {"linux": True}}, "clearpart --linux"),
    (
        {
            "clearpart": {
                "all": True,
                "drives": ["hda", "hdb"],
                "list": ["sda2", "sda3"],
                "disklabel": "some-label",
                "linux": True,
            },
        },
        "clearpart --all --drives=hda,hdb --list=sda2,sda3 --disklabel=some-label --linux",
    ),
    (
        {
            "lang": "en_US.UTF-8",
            "keyboard": "us",
            "timezone": "UTC",
            "zerombr": True,
            "clearpart": {"all": True, "drives": ["sd*|hd*|vda", "/dev/vdc"]},
        },
        "lang en_US.UTF-8\nkeyboard us\ntimezone UTC\nzerombr\nclearpart --all --drives=sd*|hd*|vda,/dev/vdc",
    ),
    ({"reboot": True}, "reboot"),
    ({"reboot": {"eject": False}}, "reboot"),
    ({"reboot": {"eject": True}}, "reboot --eject"),
    ({"reboot": {"kexec": False}}, "reboot"),
    ({"reboot": {"kexec": True}}, "reboot --kexec"),
    ({"reboot": {"eject": True, "kexec": True}}, "reboot --eject --kexec"),
    ({"display_mode": "text"}, "text"),
    ({"display_mode": "graphical"}, "graphical"),
    ({"display_mode": "cmdline"}, "cmdline"),
    # autopart
    ({"autopart": {}}, "autopart"),
    ({"autopart": {"type": "plain"}}, "autopart --type=plain"),
    ({"autopart": {"fstype": "ext4"}}, "autopart --fstype=ext4"),
    ({"autopart": {"nolvm": True}}, "autopart --nolvm"),
    ({"autopart": {"encrypted": True}}, "autopart --encrypted"),
    ({"autopart": {"passphrase": "secret"}}, "autopart --passphrase=secret"),
    ({"autopart": {"escrowcert": "http://escrow"}}, "autopart --escrowcert=http://escrow"),
    ({"autopart": {"backuppassphrase": True}}, "autopart --backuppassphrase"),
    ({"autopart": {"cipher": "aes-xts-plain2048"}}, "autopart --cipher=aes-xts-plain2048"),
    ({"autopart": {"luks-version": "42"}}, "autopart --luks-version=42"),
    ({"autopart": {"pbkdf": "scrypt"}}, "autopart --pbkdf=scrypt"),
    ({"autopart": {"pbkdf-memory": 64}}, "autopart --pbkdf-memory=64"),
    ({"autopart": {"pbkdf-time": 128}}, "autopart --pbkdf-time=128"),
    ({"autopart": {"pbkdf-iterations": 256}}, "autopart --pbkdf-iterations=256"),
    ({"autopart": {"nohome": True}}, "autopart --nohome"),
    ({"autopart": {"noswap": True}}, "autopart --noswap"),
    ({"autopart": {"type": "plain", "fstype": "xfs", "nohome": True}}, "autopart --type=plain --fstype=xfs --nohome"),
    ({"autopart": {"type": "plain", "fstype": "xfs", "nohome": True, "noswap": True}},
     "autopart --type=plain --fstype=xfs --nohome --noswap"),
    ({
        "lang": "en_US.UTF-8",
        "keyboard": "us",
        "timezone": "UTC",
        "zerombr": True,
        "clearpart": {
            "all": True,
            "drives": [
                "sd*|hd*|vda",
                "/dev/vdc"
            ]
        },
        "autopart": {
            "type": "lvm",
            "fstype": "zfs",
            "nolvm": True,
            "encrypted": True,
            "passphrase": "secret2",
            "escrowcert": "http://some-url",
            "backuppassphrase": True,
            "cipher": "twofish-cbc",
            "luks-version": "2",
            "pbkdf": "scrypt",
            "pbkdf-memory": 256,
            "pbkdf-time": 512,
            # pbkdf-iterations cannot be used together with time
        },
    },
        "lang en_US.UTF-8\nkeyboard us\ntimezone UTC\nzerombr\n" +
        "clearpart --all --drives=sd*|hd*|vda,/dev/vdc\n" +
        "autopart --type=lvm --fstype=zfs --nolvm --encrypted" +
        " --passphrase=secret2 --escrowcert=http://some-url" +
        " --backuppassphrase --cipher=twofish-cbc --luks-version=2" +
        " --pbkdf=scrypt --pbkdf-memory=256 --pbkdf-time=512"
    ),
    # network is always a list
    ({"network": [{"device": "foo", "activate": False}, {"device": "bar", "bootproto": "dhcp"}]},
     "network --device=foo --no-activate\nnetwork --device=bar --bootproto=dhcp"),
    ({"network": [{"device": "foo", "activate": True}]}, "network --device=foo --activate"),
    ({"network": [{"device": "foo", "activate": False}]}, "network --device=foo --no-activate"),
    ({"network": [{"device": "foo", "bootproto": "dhcp"}]}, "network --device=foo --bootproto=dhcp"),
    ({"network": [{"device": "foo", "onboot": "on"}]}, "network --device=foo --onboot=on"),
    ({"network": [{"device": "foo", "ip": "10.0.0.2"}]}, "network --device=foo --ip=10.0.0.2"),
    ({"network": [{"device": "foo", "ip": "auto"}]}, "network --device=foo --ip=auto"),
    ({"network": [{"device": "foo", "ipv6": "3ffe:ffff:0:1::1/128"}]},
     "network --device=foo --ipv6=3ffe:ffff:0:1::1/128"),
    ({"network": [{"device": "foo", "ipv6": "dhcp"}]}, "network --device=foo --ipv6=dhcp"),
    ({"network": [{"device": "foo", "gateway": "10.0.0.1"}]}, "network --device=foo --gateway=10.0.0.1"),
    ({"network": [{"device": "foo", "ipv6gateway": "FE80::1"}]}, "network --device=foo --ipv6gateway=FE80::1"),
    ({"network": [{"device": "foo", "nameservers": ["1.1.1.1"]}]}, "network --device=foo --nameserver=1.1.1.1"),
    ({"network": [{"device": "foo", "nameservers": ["1.1.1.1", "8.8.8.8"]}]},
     "network --device=foo --nameserver=1.1.1.1 --nameserver=8.8.8.8"),
    ({"network": [{"device": "foo", "netmask": "255.255.0.0"}]}, "network --device=foo --netmask=255.255.0.0"),
    ({"network": [{"device": "foo", "hostname": "meep"}]}, "network --device=foo --hostname=meep"),
    ({"network": [{"device": "foo", "essid": "wlan-123"}]}, "network --device=foo --essid=wlan-123"),
    ({"network": [{"device": "foo", "wpakey": "secret"}]}, "network --device=foo --wpakey=secret"),
    # device= can be written in multiple forms, see
    # https://pykickstart.readthedocs.io/en/latest/kickstart-docs.html#network
    ({"network": [{"device": "1234567890123456"}]}, "network --device=1234567890123456"),
    ({"network": [{"device": "em1"}]}, "network --device=em1"),
    ({"network": [{"device": "01:23:45:67:89:ab"}]}, "network --device=01:23:45:67:89:ab"),
    ({"network": [{"device": "link"}]}, "network --device=link"),
    ({"network": [{"device": "bootif"}]}, "network --device=bootif"),
    # hostname alone can be 63 chars (see hostname(7))
    ({"network": [{"device": "foo", "hostname": "123456789012345678901234567890123456789012345678901234567890123"}]},
     "network --device=foo --hostname=123456789012345678901234567890123456789012345678901234567890123"),
    # hostname can also be written as a QFDN
    ({"network": [{"device": "foo", "hostname": "foo.bar.com"}]},
     "network --device=foo --hostname=foo.bar.com"),
    # ostreecontainer
    (
        {
            "ostreecontainer": {
                "stateroot": "some-osname",
                "url": "http://some-ostree-url.com/foo",
                "transport": "registry",
                "remote": "some-remote",
                "signatureverification": False,
            },
        },
        "ostreecontainer --url=http://some-ostree-url.com/foo --stateroot=some-osname --transport=registry --remote=some-remote",
    ),
    (
        {
            "ostreecontainer": {
                "url": "http://some-ostree-url.com/foo",
            },
        },
        "ostreecontainer --url=http://some-ostree-url.com/foo",
    ),
    ({"ostreecontainer": {"transport": "oci", "url": "/run/install/repo/container", }, },
     "ostreecontainer --url=/run/install/repo/container --transport=oci",),
    ({"ostreecontainer": {"transport": "oci-archive", "url": "/run/install/repo/container.tar", }, },
     "ostreecontainer --url=/run/install/repo/container.tar --transport=oci-archive",),
    ({"ostreecontainer": {"transport": "dir", "url": "/run/install/repo/container", }, },
     "ostreecontainer --url=/run/install/repo/container --transport=dir",),
    ({"bootloader": {"append": "karg1 karg2=0"}}, "bootloader --append='karg1 karg2=0'"),

    # bootc
    ({"bootc": {"source-imgref": "docker://quay.io/fedora/fedora-bootc:latest"}},
     "bootc --source-imgref docker://quay.io/fedora/fedora-bootc:latest"),
    ({"bootc": {"source-imgref": "oci:/run/install/repo/container", "target-imgref": "docker://quay.io/fedora/fedora-bootc:latest"}},
     "bootc --source-imgref oci:/run/install/repo/container --target-imgref docker://quay.io/fedora/fedora-bootc:latest"),

    # %post
    ({"%post": [{"commands": ["mkdir /scratch"]}]}, "%post\nmkdir /scratch\n%end"),
    (
        {
            "%post": [
                {"commands": ["mkdir /scratch"]},
                {"commands": ["print('DONE!!!')"], "interpreter": "/usr/bin/python3"},
            ]
        },
        "%post\nmkdir /scratch\n%end\n" +
        "%post --interpreter \"/usr/bin/python3\"\nprint('DONE!!!')\n%end"
    ),
    (
        {
            "%post": [
                {"commands": ["mkdir /scratch"]},
                {
                    "erroronfail": True,
                    "nochroot": True,
                    "interpreter": "/usr/bin/bash",
                    "log": "/mnt/sysimage/var/log/ks-p2.log",
                    "commands": [
                        "echo 'Starting post2'",
                        "if [ ! -e /mnt/sysimage/etc/resolv.conf ]; then",
                        "  cp /etc/resolv.conf /mnt/sysimage/etc/resolv.conf",
                        "fi",
                    ]
                },
                {"commands": ["print('DONE!!!')"], "interpreter": "/usr/bin/python3"},
            ]
        },
        "%post\nmkdir /scratch\n%end\n" +
        "%post --erroronfail --nochroot --log \"/mnt/sysimage/var/log/ks-p2.log\" --interpreter \"/usr/bin/bash\"\n" +
        "echo 'Starting post2'\n" +
        "if [ ! -e /mnt/sysimage/etc/resolv.conf ]; then\n" +
        "  cp /etc/resolv.conf /mnt/sysimage/etc/resolv.conf\n" +
        "fi\n" +
        "%end\n" +
        "%post --interpreter \"/usr/bin/python3\"\nprint('DONE!!!')\n%end"
    )
]


STAGE_NAME = "org.osbuild.kickstart"


@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
@pytest.mark.parametrize("test_input,expected", TEST_INPUT)
def test_kickstart_test_cases_valid(stage_schema, test_input, expected):  # pylint: disable=unused-argument
    """ ensure all test inputs are valid """
    test_data = {
        "name": STAGE_NAME,
        "options": {
            "path": "some-path",
        }
    }
    test_data["options"].update(test_input)
    res = stage_schema.validate(test_data)
    assert res.valid is True, f"input: {test_input}\nerr: {[e.as_dict() for e in res.errors]}"


@pytest.mark.parametrize("test_input,expected", TEST_INPUT)
def test_kickstart_write(tmp_path, stage_module, test_input, expected):
    ks_path = "kickstart/kfs.cfg"
    options = {"path": ks_path}
    options.update(test_input)

    stage_module.main(tmp_path, options)

    ks_path = os.path.join(tmp_path, ks_path)
    with open(ks_path, encoding="utf-8") as fp:
        ks_content = fp.read()
    assert ks_content == expected + "\n"


@pytest.mark.skipif(not has_executable("ksvalidator"), reason="`ksvalidator` is required")
@pytest.mark.parametrize("test_input,expected", TEST_INPUT)
def test_kickstart_valid(tmp_path, stage_module, test_input, expected):  # pylint: disable=unused-argument
    ks_path = "kickstart/kfs.cfg"
    options = {"path": ks_path}
    options.update(test_input)

    stage_module.main(tmp_path, options)

    ks_path = os.path.join(tmp_path, ks_path)

    # check with pykickstart if the file looks valid
    subprocess.check_call(["ksvalidator", ks_path])


@pytest.mark.parametrize(
    "test_input,expected_err",
    [
        # BAD pattern, ensure some obvious ways to write arbitrary
        # kickstart files will not work
        ({"clearpart": {}}, "{} is not valid "),
        ({"clearpart": {"disklabel": r"\n%pre\necho p0wnd"}}, r"p0wnd' does not match"),
        ({"clearpart": {"drives": [" --spaces-dashes-not-allowed"]}}, "' --spaces-dashes-not-allowed' does not match"),
        ({"clearpart": {"drives": ["\n%pre not allowed"]}}, "not allowed' does not match"),
        ({"clearpart": {"drives": ["no,comma"]}}, "no,comma' does not match"),
        ({"clearpart": {"list": ["\n%pre not allowed"]}}, "not allowed' does not match"),
        ({"clearpart": {"list": ["no,comma"]}}, "no,comma' does not match"),
        ({"clearpart": {"disklabel": "\n%pre not allowed"}}, "not allowed' does not match"),
        ({"clearpart": {"random": "option"}}, "is not valid "),
        # schema ensures reboot has at least one option set
        ({"reboot": {}}, "{} is not valid under any of the given schemas"),
        ({"reboot": "random-string"}, "'random-string' is not valid "),
        ({"reboot": {"random": "option"}}, "{'random': 'option'} is not valid "),
        ({"display_mode": "invalid-mode"}, "'invalid-mode' is not one of "),
        # autopart
        ({"autopart": {"type": "not-valid"}}, r"'not-valid' is not one of \["),
        # Only one of --pbkdf-{time,iterations} can be specified at the same time
        ({"autopart": {"pbkdf-time": 1, "pbkdf-iterations": 2}}, r"( should not be valid under| is not allowed for)"),
        # network is always a list
        ({"network": {"device": "foo"}}, " is not of type 'array'"),
        ({"network": [{"device": "foo", "activate": "string"}]}, " is not of type 'boolean'"),
        ({"network": [{"device": "foo", "random": "option"}]}, "Additional properties are not allowed "),
        ({"network": [{"device": "foo", "bootproto": "invalid"}]}, r" is not one of \["),
        ({"network": [{"device": "foo", "ip": "invalid"}]}, " does not match "),
        ({"network": [{"device": "foo", "ip": "256.1.1.1"}]}, " does not match "),
        ({"network": [{"device": "foo", "ip": "1.256.1.1"}]}, " does not match "),
        ({"network": [{"device": "foo", "ip": "1.1.256.1"}]}, " does not match "),
        ({"network": [{"device": "foo", "ip": "1.1.1.256"}]}, " does not match "),
        # kernel will accept this (and make it 127.0.0.1) but it's
        # technically not valid. if this becomes a problem we may need to
        # relax (or remove) the ipv4 validation regex. Also
        # 127.256 will be accepted and overflows into "127.0.1.0".
        ({"network": [{"device": "foo", "ip": "127.1"}]}, " does not match "),
        ({"network": [{"device": "foo", "gateway": "invalid"}]}, " does not match "),
        ({"network": [{"device": "foo", "nameservers": ["invalid"]}]}, " does not match "),
        # schema says at least 2 chars (this is arbitrary)
        ({"network": [{"device": "f"}]}, " does not match "),
        # device name can be max 16 chars (see IFNAMSIZ in the kernel source)
        # but otherwise are very free from, see
        # https://elixir.bootlin.com/linux/v6.6.1/source/net/core/dev.c#L1038
        ({"network": [{"device": "12345678901234567"}]}, " does not match "),
        # when specificed via MAC address it 17 chars (12 chars plus 5 ":")
        # and look like a mac address
        ({"network": [{"device": "00:01"}]}, " does not match "),
        ({"network": [{"device": "00:XX"}]}, " does not match "),
        ({"network": [{"device": "foo/bar"}]}, " does not match "),
        ({"network": [{"device": "foo?"}]}, " does not match "),
        # see hostname(7)
        ({"network": [{"device": "foo",
                       "hostname": "1234567890123456789012345678901234567890123456789012345678901234"}]}, " does not match "),
        ({"network": [{"device": "foo", "hostname": "x$"}]}, " does not match "),
        ({"network": [{"device": "foo", "hostname": "-invalid"}]}, " does not match "),
        ({"network": [{"device": "foo", "hostname": "foo..bar"}]}, " does not match "),
        # not more than 253 chars (63*3 + 62 + 3 dots = 254)
        ({"network": [{"device": "foo",
                       "hostname":
                       "123456789012345678901234567890123456789012345678901234567890123" +
                       ".123456789012345678901234567890123456789012345678901234567890123" +
                       ".123456789012345678901234567890123456789012345678901234567890123" +
                       ".12345678901234567890123456789012345678901234567890123456789012"
                       }]}, " does not match "),
        # ostreecontainer
        ({"ostreecontainer": {"url": "http://some-ostree-url.com/foo",
         "transport": "not-valid"}}, r"'not-valid' is not one of \["),
        # not both ostreecontainer and ostree
        (
            {
                "ostreecontainer": {
                    "url": "http://some-ostree-url.com/foo",
                },
                "ostree": {
                    "osname": "some-osname",
                    "url": "http://some-ostree-url.com/foo",
                    "ref": "some-ref",
                    "remote": "some-remote",
                    "gpg": True,
                },
            },
            "is valid under each of",
        ),
        # not both bootc and ostree
        (
            {
                "bootc": {
                    "source-imgref": "docker://quay.io/fedora/fedora-bootc:latest",
                },
                "ostree": {
                    "osname": "some-osname",
                    "url": "http://some-ostree-url.com/foo",
                    "ref": "some-ref",
                },
            },
            "is valid under each of",
        ),
        # not both bootc and ostreecontainer
        (
            {
                "bootc": {
                    "source-imgref": "docker://quay.io/fedora/fedora-bootc:latest",
                },
                "ostreecontainer": {
                    "url": "http://some-ostree-url.com/foo",
                },
            },
            "is valid under each of",
        ),
        # bootc requires source-imgref
        ({"bootc": {}}, "'source-imgref' is a required property"),
        ({"rootpw": {}}, "is not valid under any of the given schemas"),
        ({"rootpw": {"lock": True, "allow_ssh": True}}, "is not valid under any of the given schemas"),
        ({"rootpw": {"plaintext": True}}, "is not valid under any of the given schemas"),
        # under py3.6 the message is "'' is too short" under other versions the message is "'' should be non-empty"
        ({"rootpw": {"plaintext": True, "password": ""}}, re.compile("'' should be non-empty|'' is too short")),
        ({"rootpw": {"iscrypted": True}}, "is not valid under any of the given schemas"),
        ({"rootpw": {"password": "password"}}, "is not valid under any of the given schemas"),
        (
            {"rootpw": {"iscrypted": True, "plaintext": True, "password": "pass"}},
            "is not valid under any of the given schemas"
        ),
        # bad %post blocks
        ({"%post": []}, re.compile(r"\[\] should be non-empty|\[\] is too short")),
        ({"%post": [{}]}, "'commands' is a required property"),
        ({"%post": [{"commands": []}]}, re.compile(r"\[\] should be non-empty|\[\] is too short")),
    ],
)
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation_bad_apples(stage_schema, test_input, expected_err):
    test_data = {
        "name": STAGE_NAME,
        "options": {
            "path": "some-path",
        }
    }
    test_data["options"].update(test_input)
    res = stage_schema.validate(test_data)

    assert res.valid is False
    testutil.assert_jsonschema_error_contains(res, re.compile(expected_err), expected_num_errs=1)



#!/usr/bin/python3

import os
import subprocess

import pytest  # type: ignore

from osbuild import testutil
from osbuild.testutil import has_executable

STAGE_NAME = "org.osbuild.dmverity"


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, "'root_hash_file' is a required property"),
    ({"root_hash_file": 123}, "123 is not of type 'string'"),
    # good
    ({"root_hash_file": "abc"}, ""),
])
def test_schema_validation_dmverity(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "devices": {
            "data_device": {
                "path": "some-path",
            },
            "hash_device": {
                "path": "some-path",
            },
        },
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.skipif(not has_executable("mkfs.ext4"), reason="need mkfs.ext4")
@pytest.mark.skipif(not has_executable("veritysetup"), reason="need veritysetup")
def test_dmverity_integration(tmp_path, stage_module):
    fake_dev_path = tmp_path / "dev"
    fake_dev_path.mkdir()

    fake_data_disk = "xxd1"
    fake_hash_disk = "xxd2"
    for fname in [fake_data_disk, fake_hash_disk]:
        p = fake_dev_path / fname
        p.write_bytes(b"")
        os.truncate(p, 10 * 1024 * 1024)
    # format is not strictly needed as dmvertify is working on the block level but this makes the test more realistic
    subprocess.run(
        ["mkfs.ext4", os.fspath(fake_dev_path / fake_data_disk)], check=True)

    paths = {
        "devices": fake_dev_path,
    }
    devices = {
        "data_device": {
            "path": fake_data_disk,
        },
        "hash_device": {
            "path": fake_hash_disk,
        },
    }
    options = {
        "root_hash_file": "hashfile",
    }

    tree = tmp_path
    stage_module.main(tree, paths, devices, options)
    output = subprocess.check_output(
        ["veritysetup", "dump", os.fspath(fake_dev_path / fake_hash_disk)],
        universal_newlines=True)
    assert "UUID:" in output
    # hash file is created and has the expected size
    assert (tree / "hashfile").stat().st_size == 64



#!/usr/bin/python3

import os
import pathlib
from unittest.mock import patch

import pytest

import osbuild.meta
from osbuild import testutil

MOUNTS_NAME = "org.osbuild.bind"


@patch("subprocess.run")
def test_bind_mount_simple(mocked_run, mounts_service):
    fake_tree_path = "/tmp/osbuild/.osbuild/tmp/buildroot-tmp-_ym0rt5a/mounts/"
    fake_mountroot_path = "/tmp/osbuild/.osbuild/stage/uuid-e7443c948dff406d88bcbb7658a31f0f/data/tree/"

    fake_args = {
        "tree": fake_tree_path,
        "root": fake_mountroot_path,
        "target": "tree://",
        "options": {
            "source": "mount://",
        }
    }
    mounts_service.mount(fake_args)
    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert args[0] == ["mount", "--rbind", fake_mountroot_path, fake_tree_path]
    assert kwargs == {"check": True}


@patch("subprocess.run")
def test_bind_umount_simple(mocked_run, mounts_service):
    # nothing mounted yet, nothing is done
    mounts_service.umount()
    assert len(mocked_run.call_args_list) == 0
    # pretend a mount
    mounts_service.mountpoint = "/something"
    mounts_service.umount()
    assert len(mocked_run.call_args_list) == 1
    args, kwargs = mocked_run.call_args_list[0]
    assert args[0] == ["umount", "-R", "-v", "/something"]
    assert kwargs == {"check": True}
    # mountpoint is cleared
    assert mounts_service.mountpoint == ""
    # ensure we do not umount twice
    mounts_service.umount()
    assert len(mocked_run.call_args_list) == 1


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, "'name' is a required property"),
    ({}, "'target' is a required property"),
    ({}, "'source' is a required property"),
    # good
    ({"name": "bind", "target": "tree://", "options": {"source": "mount://"}}, ""),
])
def test_parameters_validation(test_data, expected_err):
    root = pathlib.Path(__file__).parent.parent.parent
    mod_info = osbuild.meta.ModuleInfo.load(root, "Mount", MOUNTS_NAME)
    schema = osbuild.meta.Schema(mod_info.get_schema(), MOUNTS_NAME)
    test_input = {
        "type": MOUNTS_NAME,
        "options": {}
    }
    test_input.update(test_data)
    res = schema.validate(test_input)
    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err)


@pytest.mark.skipif(os.getuid() != 0, reason="needs root")
def test_bind_mount_integration(tmp_path, mounts_service):
    fake_tree_path = tmp_path / "mounts"
    fake_tree_path.mkdir(parents=True, exist_ok=True)

    fake_mountroot_path = tmp_path / "uuid-1234/data/tree"
    fake_mountroot_path.mkdir(parents=True, exist_ok=True)
    (fake_mountroot_path / "in-src-mnt").write_text("", encoding="utf8")

    fake_args = {
        "tree": fake_tree_path,
        "root": fake_mountroot_path,
        "target": "tree://",
        "options": {
            "source": "mount://",
        }
    }
    assert not (fake_tree_path / "in-src-mnt").exists()
    mounts_service.mount(fake_args)
    assert (fake_tree_path / "in-src-mnt").exists()
    mounts_service.umount()
    assert not (fake_tree_path / "in-src-mnt").exists()



#!/usr/bin/python3

import os
import re

import pytest

from osbuild.testutil import assert_jsonschema_error_contains

STAGE_NAME = "org.osbuild.ln"


def test_ln_symbolic(tmp_path, stage_module):
    tree = tmp_path / "tree"
    tree.mkdir()
    (tree / "target1").touch()
    (tree / "target2").touch()
    (tree / "dir").mkdir()
    (tree / "dir" / "target3").touch()

    options = {
        "paths": [
            {
                "target": "target1",
                "link_name": "tree:///link1",
                "symbolic": True
            },
            {
                "target": "target2",
                "link_name": "tree:///dir/link2",
                "symbolic": True
            },
            {
                "target": "/etc/os-release",
                "link_name": "tree:///os-release",
                "symbolic": True
            }
        ],
    }

    args = {
        "tree": os.fspath(tree)
    }

    stage_module.main(args, options)

    assert os.path.islink(tree / "link1")
    assert os.readlink(tree / "link1") == "target1"
    assert os.path.islink(tree / "dir/link2")
    assert os.readlink(tree / "dir/link2") == "target2"
    assert os.path.islink(tree / "os-release")
    assert os.readlink(tree / "os-release") == "/etc/os-release"


def test_ln_hardlink(tmp_path, stage_module):
    tree = tmp_path / "tree"
    tree.mkdir()
    target_path = tree / "target"
    target_path.write_text("content")

    options = {
        "paths": [
            {
                "target": "target",
                "link_name": "tree:///link",
                "symbolic": False
            }
        ],
    }

    args = {
        "tree": os.fspath(tree)
    }

    old_cwd = os.getcwd()
    os.chdir(tree)
    try:
        stage_module.main(args, options)
    finally:
        os.chdir(old_cwd)

    link_path = tree / "link"
    assert link_path.exists()
    assert not link_path.is_symlink()
    assert target_path.stat().st_ino == link_path.stat().st_ino


@pytest.mark.parametrize("test_data,expected_err", [
    # bad
    ({}, "'paths' is a required property"),
    ({"paths": [{"link_name": "a", "target": "b"}]}, "is not valid under any of the given schemas"),
    ({"paths": "not-an-array"}, "'not-an-array' is not of type 'array'"),
    ({"paths": []}, re.compile(r"\[\] should be non-empty|\[\] is too short")),
    ({"paths": [{"target": "b"}]}, "'link_name' is a required property",),
    ({"paths": [{"link_name": "tree:///a"}]}, "'target' is a required property"),
    ({"paths": [{"link_name": "tree:///a", "target": "b", "symbolic": "not-a-bool"}]},
     "'not-a-bool' is not of type 'boolean'"),
    # good
    ({"paths": [{"link_name": "tree:///a", "target": "b"}]}, ""),
    ({"paths": [{"link_name": "tree:///a", "target": "b", "symbolic": True}]}, ""),
    ({"paths": [{"link_name": "mount://root/a", "target": "b", "symbolic": False}]}, ""),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)



#!/usr/bin/python3

import os.path
import re

import pytest

from osbuild import testutil
from osbuild.util import toml

TEST_INPUT = [
    ({}, {"additionalimagestores": ["/path/to/store"]}, [("storage.options.additionalimagestores", ["/path/to/store"])]),
    ({}, {"additionalimagestores": ["/path/to/store"]}, [("storage.options.additionalimagestores", ["/path/to/store"])]),
    ({"transient_store": True, "runroot": "/foo"}, {}, [("storage.transient_store", True), ("storage.runroot", '/foo')]),
]


STAGE_NAME = "org.osbuild.containers.storage.conf"


@pytest.mark.tomlwrite
@pytest.mark.parametrize("test_filename", ["/etc/containers/storage.conf", "/usr/share/containers/storage.conf"])
@pytest.mark.parametrize("test_storage,test_options,expected", TEST_INPUT)
def test_containers_storage_conf_integration(tmp_path, stage_module, test_filename, test_storage, test_options, expected):  # pylint: disable=unused-argument
    treedir = os.path.join(tmp_path, "tree")
    confdir, confname = os.path.split(test_filename.lstrip("/"))
    confdir = os.path.join(treedir, confdir)
    os.makedirs(confdir, exist_ok=True)

    options = {
        "filename": test_filename,
        "config": {
            "storage": {
                "options": {
                }
            }
        }
    }

    options["config"]["storage"].update(test_storage)
    options["config"]["storage"]["options"].update(test_options)

    stage_module.main(treedir, options)

    confpath = os.path.join(confdir, confname)
    assert os.path.exists(confpath)

    conf = None
    conf = toml.load_from_file(confpath)

    assert conf is not None

    for (key, value) in expected:
        testutil.assert_dict_has(conf, key, value)


@pytest.mark.parametrize(
    "test_data,storage_test_data,expected_err",
    [
        # None, note that starting from jsonschema 4.21.0 the error changes
        # so we need a regexp here
        ({}, {}, re.compile("does not have enough properties|should be non-empty")),
        # All options
        ({
            "filename": "/etc/containers/storage.conf",
            "filebase": "/some/path",
            "comment": ["List of", "comments"]
        },
            {
            "driver": "vfs",
            "graphroot": "/some/path",
            "runroot": "/other/path",
            "transient_store": True,
            "options": {
                "additionalimagestores": ["/path/to/store"],
                "pull_options": {
                    "enable_partial_images": "true",
                    "use_hard_links": "false",
                    "convert_images": "false"
                },
                "overlay": {
                    "mountopt": "metadata"
                }
            }
        }, ""),
        ({}, {"transient_store": "true"}, "is not of type 'boolean'"),
        # Pull options are strings, not booleans (in the config file format)
        ({}, {"options": {"pull_options": {"use_hard_links": "foobar"}}}, "is not one of"),
        ({}, {"options": {"pull_options": {"use_hard_links": True}}}, "is not one of"),
        ({}, {"options": {"pull_options": {"use_hard_links": True}}}, "is not of type 'string'"),
    ],
)
def test_schema_validation_containers_storage_conf(stage_schema, test_data, storage_test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {
            "config": {
                "storage": {
                }
            }
        }
    }

    test_input["options"].update(test_data)
    test_input["options"]["config"]["storage"].update(storage_test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err)



"""Red Hat Subscription Manager support module

This module implements utilities that help with interactions
with the subscriptions attached to the host machine.
"""

import configparser
import contextlib
import glob
import os
import re
from typing import List


class Subscriptions:
    DEFAULT_SSL_CA_CERT = "/etc/rhsm/ca/redhat-uep.pem"
    DEFAULT_ENTITLEMENT_DIR = "/etc/pki/entitlement"
    DEFAULT_REPO_FILE = "/etc/yum.repos.d/redhat.repo"

    def __init__(self, repositories):
        self.repositories = repositories
        # These are used as a fallback if the repositories don't
        # contain secrets for a requested URL.
        self.secrets = None

        if self.is_container_with_rhsm_secrets():
            self.DEFAULT_SSL_CA_CERT = "/run/secrets/rhsm/ca/redhat-uep.pem"
            self.DEFAULT_ENTITLEMENT_DIR = "/run/secrets/etc-pki-entitlement"
            self.DEFAULT_REPO_FILE = "/run/secrets/redhat.repo"

    @staticmethod
    def is_container_with_rhsm_secrets():
        """Detect if we are running inside a podman container and RHSM secrets are available."""
        return os.path.exists("/run/.containerenv") and os.path.exists("/run/secrets")

    def get_fallback_rhsm_secrets(self):
        rhsm_secrets = {
            'ssl_ca_cert': self.DEFAULT_SSL_CA_CERT,
            'ssl_client_key': "",
            'ssl_client_cert': ""
        }

        keys = glob.glob(f"{self.DEFAULT_ENTITLEMENT_DIR}/*-key.pem")
        for key in keys:
            # The key and cert have the same prefix
            cert = key.rstrip("-key.pem") + ".pem"
            # The key is only valid if it has a matching cert
            if os.path.exists(cert):
                rhsm_secrets['ssl_client_key'] = key
                rhsm_secrets['ssl_client_cert'] = cert
                # Once the dictionary is complete, assign it to the object
                self.secrets = rhsm_secrets

        raise RuntimeError("no matching rhsm key and cert")

    @staticmethod
    def get_consumer_secrets():
        """Returns the consumer identity certificate which uniquely identifies the system.

        Will fail when running in a container. Used by ostree.
        """
        key = "/etc/pki/consumer/key.pem"
        cert = "/etc/pki/consumer/cert.pem"

        if not (os.path.exists(key) and os.path.exists(cert)):
            raise RuntimeError("rhsm consumer key and cert not found")

        return {
            'consumer_key': key,
            'consumer_cert': cert
        }

    @classmethod
    def from_host_system(cls):
        """Read redhat.repo file and process the list of repositories in there."""
        ret = cls(None)
        with contextlib.suppress(FileNotFoundError):
            with open(cls.DEFAULT_REPO_FILE, "r", encoding="utf8") as fp:
                ret = cls.parse_repo_file(fp)

        with contextlib.suppress(RuntimeError):
            ret.get_fallback_rhsm_secrets()

        if not ret.repositories and not ret.secrets:
            raise RuntimeError("No RHSM secrets found on this host.")

        return ret

    @staticmethod
    def _process_baseurl(input_url):
        """Create a regex from a baseurl.

        The osbuild manifest format does not contain information about repositories.
        It only includes URLs of each RPM. In order to make this RHSM support work,
        osbuild needs to find a relation between a "baseurl" in a *.repo file and the
        URL given in the manifest. To do so, it creates a regex from all baseurls
        found in the *.repo file and matches them against the URL.
        """
        # First escape meta characters that might occur in a URL
        input_url = re.escape(input_url)

        # Now replace variables with regexes (see man 5 yum.conf for the list)
        for variable in ["\\$releasever", "\\$arch", "\\$basearch", "\\$uuid"]:
            input_url = input_url.replace(variable, "[^/]*")

        return re.compile(input_url)

    @classmethod
    def parse_repo_file(cls, fp):
        """Take a file object and reads its content assuming it is a .repo file."""
        parser = configparser.ConfigParser(interpolation=None)
        parser.read_file(fp)

        repositories = {}
        for section in parser.sections():
            current = {
                "matchurl": cls._process_baseurl(parser.get(section, "baseurl"))
            }

            # On RHEL systems registered to Insights, the redhat.repo exists and contains entitlement
            # certificates, however, "sslcacert" is unset and rhsm dnf plugin automatically sets that
            # to /etc/rhsm/ca/redhat-uep.pem or /run/secrets/rhsm/ca/redhat-uep.pem respectively.
            current["sslcacert"] = parser.get(section, "sslcacert", fallback=cls.DEFAULT_SSL_CA_CERT)

            # Entitlement certificates are always present in redhat.repo
            for parameter in ["sslclientkey", "sslclientcert"]:
                current[parameter] = parser.get(section, parameter)

            repositories[section] = current

        return cls(repositories)

    def get_secrets(self, urls: List[str]):
        """
        Get the RHSM secrets for a list of URLs.

        The URLs can be a baseurl, metalink, or mirrorlist. The list can
        even combine all types of URLs. The function will iterate over the list
        and try to find a matching URL in the redhat.repo file. The function
        returns secrets for the first matching URL.

        If no matching URL is found, the function will try the fallback
        secrets. If no fallback secrets are found, the function will raise
        a RuntimeError.
        """
        # Try to find a matching URL from redhat.repo file first
        if self.repositories is not None:
            for parameters in self.repositories.values():
                for url in urls:
                    if parameters["matchurl"].match(url) is not None:
                        return {
                            "ssl_ca_cert": parameters["sslcacert"],
                            "ssl_client_key": parameters["sslclientkey"],
                            "ssl_client_cert": parameters["sslclientcert"]
                        }

        # In case there is no matching URL, try the fallback
        if self.secrets:
            return self.secrets

        raise RuntimeError(f"There are no RHSM secret associated with {urls}")



#!/usr/bin/python3

import os
import subprocess
import textwrap

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.systemd.unit.create"


@pytest.mark.parametrize("test_data,expected_err", [
    # good
    (
        {
            "filename": "foo.service",
            "config": {
                "Unit": {},
                "Service": {},
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.mount",
            "config": {
                "Unit": {},
                "Mount": {"What": "", "Where": ""},
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.socket",
            "config": {
                "Socket": {"ListenStream": "/run/test/api.socket"},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.swap",
            "config": {
                "Swap": {"What": ""},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "inherit",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "null",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "tty",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "journal",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "kmsg",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "journal+console",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "kmsg+console",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "file:/var/log/example.log",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "append:/var/log/app.log",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "append:/root/important.txt",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "truncate:/debug.log",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "socket",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "fd:stdout",
                },
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "fd:whatever",
                },
                "Install": {},
            },
        },
        "",
    ),

    # bad
    # # No filename
    ({"config": {"Unit": {}, "Service": {}, "Install": {}}}, "'filename' is a required property"),

    # # no config block
    ({"filename": "foo.service"}, "'config' is a required property"),

    # # no Unit section in .service
    ({"filename": "foo.service", "config": {"Service": {}, "Install": {}}},
     "{'Service': {}, 'Install': {}} is not valid under any of the given schemas"),

    # # no Service section in .service
    ({"filename": "foo.service", "config": {"Unit": {}, "Install": {}}},
     "{'Unit': {}, 'Install': {}} is not valid under any of the given schemas"),

    # # no Install section in .service
    ({"filename": "foo.service", "config": {"Unit": {}, "Service": {}}},
     "{'Unit': {}, 'Service': {}} is not valid under any of the given schemas"),

    # # .mount unit without Where
    ({"filename": "foo.mount", "config": {"Unit": {}, "Mount": {"What": ""}, "Install": {}}},
     "'Where' is a required property"),

    # # .mount unit without What
    ({"filename": "foo.mount", "config": {"Unit": {}, "Mount": {"Where": ""}, "Install": {}}},
     "'What' is a required property"),

    # # .service unit with Mount section
    ({"filename": "foo.service", "config": {"Unit": {}, "Service": {},
                                            "Mount": {"What": "", "Where": ""}, "Install": {}}},
     "{'Unit': {}, 'Service': {}, 'Mount': {'What': '', 'Where': ''}, "
     "'Install': {}} is not valid under any of the given schemas"),

    # # .swap unit with Service section
    ({"filename": "foo.swap", "config": {"Unit": {}, "Service": {}, "Swap": {"What": ""}, "Install": {}}},
     "{'Unit': {}, 'Service': {}, 'Swap': {'What': ''}, "
     "'Install': {}} is not valid under any of the given schemas"),

    # # bad StandardOutput values
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "syslog",
                },
                "Install": {},
            },
        },
        "'syslog' does not match",
    ),
    (
        {
            "filename": "stdout-test.service",
            "config": {
                "Unit": {},
                "Service": {
                    "StandardOutput": "file:",
                },
                "Install": {},
            },
        },
        "'file:' does not match",
    ),
])
@pytest.mark.parametrize("stage_schema", ["1"], indirect=True)
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "name": STAGE_NAME,
        "options": test_data,
    }
    res = stage_schema.validate(test_input)
    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.parametrize("test_data,expected_err", [
    # good
    (
        {
            "filename": "foo.service",
            "config": {
                "Unit": {},
                "Service": {},
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.mount",
            "config": {
                "Unit": {},
                "Mount": {"What": "", "Where": ""},
                "Install": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.socket",
            "config": {
                "Socket": {},
            },
        },
        "",
    ),
    (
        {
            "filename": "foo.swap",
            "config": {
                "Swap": {},
            },
        },
        "",
    ),
    # bad
    ({"filename": "something.service", "config": {"Unit": {}, "Mount": {}, "Install": {}}},
     "Error: something.service unit requires Service section"),
    ({"filename": "data-gifs-cats.mount", "config": {"Unit": {}, "Service": {}, "Install": {}}},
     "Error: data-gifs-cats.mount unit requires Mount section"),
    ({"filename": "data-gifs-cats.socket", "config": {"Unit": {}, "Service": {}, "Install": {}}},
     "Error: data-gifs-cats.socket unit requires Socket section"),
    ({"filename": "data-gifs-cats.swap", "config": {"Unit": {}, "Service": {}, "Install": {}}},
     "Error: data-gifs-cats.swap unit requires Swap section"),
])
def test_name_config_match(tmp_path, stage_module, test_data, expected_err):
    expected_unit_path = tmp_path / "usr/lib/systemd/system"
    expected_unit_path.mkdir(parents=True, exist_ok=True)
    try:
        stage_module.main(tmp_path, test_data)
    except ValueError as ve:
        assert expected_err == str(ve)


@pytest.mark.parametrize("unit_type,unit_path,expected_prefix", [
    ("system", "usr", "usr/lib/systemd/system"),
    ("system", "etc", "etc/systemd/system"),
    ("global", "usr", "usr/lib/systemd/user"),
    ("global", "etc", "etc/systemd/user"),
])
def test_systemd_unit_create(tmp_path, stage_module, unit_type, unit_path, expected_prefix):
    options = {
        "filename": "create-directory.service",
        "unit-type": unit_type,
        "unit-path": unit_path,
        "config": {
            "Unit": {
                "Description": "Create directory",
                "DefaultDependencies": False,
                "ConditionPathExists": [
                    "|!/etc/myfile"
                ],
                "ConditionPathIsDirectory": [
                    "|!/etc/mydir"
                ],
                # need to use real units otherwise the validation will fail
                "Wants": [
                    "basic.target",
                    "sysinit.target",
                ],
                "Requires": [
                    "paths.target",
                    "sockets.target",
                ],
                "After": [
                    "sysinit.target",
                    "default.target",
                ],
                "Before": [
                    "multi-user.target",
                ],
            },
            "Service": {
                "Type": "oneshot",
                "RemainAfterExit": True,
                "ExecStart": [
                    "/usr/bin/mkdir -p /etc/mydir",
                    "/usr/bin/touch /etc/myfile"
                ],
                "Environment": [
                    {
                        "key": "DEBUG",
                        "value": "1",
                    },
                    {
                        "key": "TRACE",
                        "value": "1",
                    },
                ],
                "EnvironmentFile": [
                    "/etc/example.env",
                    "/etc/second.env",
                ],
                "StandardOutput": "append:/var/log/mkdir.log",
            },
            "Install": {
                "WantedBy": [
                    "local-fs.target"
                ],
                "RequiredBy": [
                    "multi-user.target"
                ]
            }
        }
    }

    # create units in the tree for systemd-analyze verify
    systemd_dir = "usr/lib/systemd/system/"
    # if a unit file is empty, it gets detected as "masked" and fails verification
    testutil.make_fake_tree(tmp_path, {
        systemd_dir + "basic.target": "[Unit]\n",
        systemd_dir + "sysinit.target": "[Unit]\n",
        systemd_dir + "paths.target": "[Unit]\n",
        systemd_dir + "sockets.target": "[Unit]\n",
        "usr/bin/mkdir": "fake-mkdir",
        "usr/bin/touch": "fake-touch",
    })
    for bin_name in ["mkdir", "touch"]:
        os.chmod(tmp_path / "usr/bin" / bin_name, mode=0o755)

    expected_unit_path = tmp_path / expected_prefix / "create-directory.service"
    expected_unit_path.parent.mkdir(parents=True, exist_ok=True)

    stage_module.main(tmp_path, options)
    assert os.path.exists(expected_unit_path)
    if unit_type == "system":
        if os.environ.get("RPM_BUILD_ROOT"):
            pytest.skip("systemd-analyze can't be run in the RPM build root environment")
        # When verifying user units, systemd expects runtime directories and more generally a booted system. Setting
        # something up like that to verify it is more trouble than it's worth, especially since the system units are
        # identical.
        subprocess.run(["systemd-analyze", "verify",
                        "--man=no",  # disable manpage checks
                        f"--root={tmp_path}",  # verify commands and units in the root tree
                        expected_unit_path],
                       check=True)
    assert expected_unit_path.read_text(encoding="utf-8") == textwrap.dedent("""\
    [Unit]
    Description=Create directory
    DefaultDependencies=False
    ConditionPathExists=|!/etc/myfile
    ConditionPathIsDirectory=|!/etc/mydir
    Wants=basic.target
    Wants=sysinit.target
    Requires=paths.target
    Requires=sockets.target
    After=sysinit.target
    After=default.target
    Before=multi-user.target

    [Service]
    Type=oneshot
    RemainAfterExit=True
    ExecStart=/usr/bin/mkdir -p /etc/mydir
    ExecStart=/usr/bin/touch /etc/myfile
    Environment="DEBUG=1"
    Environment="TRACE=1"
    EnvironmentFile=/etc/example.env
    EnvironmentFile=/etc/second.env
    StandardOutput=append:/var/log/mkdir.log

    [Install]
    WantedBy=local-fs.target
    RequiredBy=multi-user.target

    """)


@pytest.mark.parametrize("unit_type,unit_path,expected_prefix", [
    ("system", "usr", "usr/lib/systemd/system"),
    ("system", "etc", "etc/systemd/system"),
    ("global", "usr", "usr/lib/systemd/user"),
    ("global", "etc", "etc/systemd/user"),
])
def test_systemd_unit_create_mount(tmp_path, stage_module, unit_type, unit_path, expected_prefix):
    options = {
        "filename": "data-test.mount",
        "unit-type": unit_type,
        "unit-path": unit_path,
        "config": {
            "Unit": {
                "DefaultDependencies": True,
                "ConditionPathExists": [
                    "/data"
                ],
                "ConditionPathIsDirectory": [
                    "/data"
                ],
                "Before": [
                    "local-fs.target",
                ],
            },
            "Mount": {
                "What": "/dev/sda2",
                "Where": "/data/test",
                "Type": "xfs",
                "Options": "rw,noatime"
            }
        }
    }

    # create units in the tree for systemd-analyze verify
    systemd_dir = "usr/lib/systemd/system/"
    # if a unit file is empty, it gets detected as "masked" and fails verification
    testutil.make_fake_tree(tmp_path, {
        systemd_dir + "basic.target": "[Unit]\n",
        systemd_dir + "sysinit.target": "[Unit]\n",
        systemd_dir + "paths.target": "[Unit]\n",
        systemd_dir + "sockets.target": "[Unit]\n",
        "usr/bin/mkdir": "fake-mkdir",
        "usr/bin/touch": "fake-touch",
    })
    for bin_name in ["mkdir", "touch"]:
        os.chmod(tmp_path / "usr/bin" / bin_name, mode=0o755)

    expected_unit_path = tmp_path / expected_prefix / "data-test.mount"
    expected_unit_path.parent.mkdir(parents=True, exist_ok=True)

    stage_module.main(tmp_path, options)
    assert os.path.exists(expected_unit_path)
    if unit_type == "system":
        if os.environ.get("RPM_BUILD_ROOT"):
            pytest.skip("systemd-analyze can't be run in the RPM build root environment")
        # When verifying user units, systemd expects runtime directories and more generally a booted system. Setting
        # something up like that to verify it is more trouble than it's worth, especially since the system units are
        # identical.
        subprocess.run(["systemd-analyze", "verify",
                        "--man=no",  # disable manpage checks
                        f"--root={tmp_path}",  # verify commands and units in the root tree
                        expected_unit_path],
                       check=True)
    assert expected_unit_path.read_text(encoding="utf-8") == textwrap.dedent("""\
    [Unit]
    DefaultDependencies=True
    ConditionPathExists=/data
    ConditionPathIsDirectory=/data
    Before=local-fs.target

    [Mount]
    What=/dev/sda2
    Where=/data/test
    Type=xfs
    Options=rw,noatime

    """)



import pytest

DEVICES_NAME = "org.osbuild.lvm2.lv"


@pytest.mark.parametrize("parent,options,expected_parent_path", [
    ("loop2", {}, "/dev/loop2"),
    ("loop1", {"vg_partnum": 2}, "/dev/loop1p2"),
])
def test_lvm2_lv_get_parent_path(devices_module, parent, options, expected_parent_path):
    pp = devices_module.get_parent_path(parent, options)
    assert pp == expected_parent_path


def test_lvm2_escaped_lv_mapper_name(devices_module):
    expected = "1d2b2150--8de2--4b68--b387--f9bc709190e8-lvroot"
    assert devices_module.escaped_lv_mapper_name("1d2b2150-8de2-4b68-b387-f9bc709190e8", "lvroot") == expected



#!/usr/bin/python3

import os.path
import subprocess
from unittest import mock

import pytest

from osbuild import testutil
from osbuild.testutil import has_executable

STAGE_NAME = "org.osbuild.mkfs.fat"


# Prepare dataset containing good and bad API call parameters
@pytest.mark.parametrize("test_data, expected_err", [
    # Bad API parameters
    ({}, "'volid' is a required property"),
    ({"volid": 123}, "123 is not of type 'string'"),
    ({"volid": "2e24ec82", "label": "12345678901234567"}, "is too long"),
    ({"volid": "2e24ec82", "fat-size": 11}, "11 is not one of [12, 16, 32]"),
    ({"volid": "2e24ec82", "geometry": "text"}, "'text' is not of type 'object'"),
    # Good API parameters
    ({"volid": "2e24ec82", "label": "12345678901"}, ""),
    ({"volid": "2e24ec82", "label": "label", "fat-size": 32}, ""),
    # mkfs requires valid volume ID, but API will accept any string right now.
    # Bellow will fail when API will be more restricted in the future.
    ({"volid": "volumeid"}, ""),
])
# This test validates only API calls using correct and incorrect queries
def test_schema_validation_mkfs_fat(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "devices": {
            "device": {
                "path": "some-path",
            },
        },
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


# This test creates dummy filesystem using predefined parameters and verifies
# that end result has those parameters set
@pytest.mark.skipif(not has_executable("mkfs.fat"), reason="need mkfs.fat")
def test_mkfs_fat_integration(tmp_path, stage_module):
    fake_disk_path = tmp_path / "fake.img"
    with fake_disk_path.open("w") as fp:
        fp.truncate(10 * 1024 * 1024)
    devices = {
        "device": {
            "path": fake_disk_path,
        },
    }
    fake_volid = "2e24ec82"
    fake_label = "osbuild"
    fat_size = 32
    options = {
        "volid": fake_volid,
        "label": fake_label,
        "fat-size": fat_size,
    }
    stage_module.main(devices, options)
    assert os.path.exists(fake_disk_path)
    output = subprocess.check_output([
        "file", "-s", fake_disk_path], encoding="utf-8")
    assert f'serial number 0x{fake_volid}' in output, \
        f'expected serial number (volid) not found in: {output}'
    assert f'label: "{fake_label}    "' in output, \
        f'expected label not found in: {output}'
    assert f'FAT ({fat_size} bit)' in output, \
        f'expected FAT size not found in: {output}'


@pytest.mark.parametrize("test_input, expected", [
    ({}, []),
    ({"fat-size": 32}, ["-F", "32"]),
    ({"fat-size": 16}, ["-F", "16"]),
    ({"fat-size": 12}, ["-F", "12"]),
    ({"label": "osbuild"}, ["-n", "osbuild"]),
    ({"geometry": {"heads": 255, "sectors-per-track": 63}}, ["-g", "255/63"]),
])
@mock.patch("subprocess.run")
def test_mkfs_fat_cmdline(mock_run, stage_module, test_input, expected):
    fake_disk_path = "/dev/xxd1"
    devices = {
        "device": {
            "path": fake_disk_path,
        },
    }
    fake_volid = "2e24ec82"
    options = {
        "volid": fake_volid,
    }
    options.update(test_input)
    stage_module.main(devices, options)

    expected = [
        "mkfs.fat",
        "-I",
        "-i", fake_volid,
    ] + expected + [
        fake_disk_path,
    ]
    mock_run.assert_called_once_with(expected, encoding="utf8", check=True)
